""" "
AlgorithmNodeFactory
"""

from dataclasses import dataclass
from typing import Any

import chipiron.environments.chess_env.board as board_mod
import chipiron.players.move_selector.treevalue.indices.node_indices as node_indices
import chipiron.players.move_selector.treevalue.nodes as node
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.boardevaluators.neural_networks.input_converters.board_representation import (
    BoardRepresentation,
    Representation364,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.factory import (
    RepresentationFactory,
)
from chipiron.players.move_selector.treevalue.node_factory.base import Base
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.node_minmax_evaluation import (
    NodeMinmaxEvaluation,
)
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode


@dataclass
class AlgorithmNodeFactory:
    """
    The classe creating Algorithm Nodes
    """

    tree_node_factory: Base[Any]
    board_representation_factory: RepresentationFactory[Any] | None
    exploration_index_data_create: node_indices.ExplorationIndexDataFactory

    def create(
        self,
        board: board_mod.IBoard,
        half_move: int,
        count: int,
        parent_node: ITreeNode | None,
        move_from_parent: moveKey | None,
        modifications: board_mod.BoardModificationP | None,
    ) -> AlgorithmNode:
        """
        Creates an AlgorithmNode object.

        Args:
            move_from_parent (chess.Move | None): the move that led to the node from the parent node
            board: The board object.
            half_move: The half move count.
            count: The count.
            parent_node: The parent node object.
            modifications: The board modifications object.

        Returns:
            An AlgorithmNode object.

        """
        tree_node: node.TreeNode[Any] = self.tree_node_factory.create(
            board=board,
            half_move=half_move,
            count=count,
            move_from_parent=move_from_parent,
            parent_node=parent_node,
            modifications=modifications,
        )
        minmax_evaluation: NodeMinmaxEvaluation = NodeMinmaxEvaluation(
            tree_node=tree_node
        )

        exploration_index_data: node_indices.NodeExplorationData | None = (
            self.exploration_index_data_create(tree_node)
        )

        board_representation: BoardRepresentation | None = None
        if self.board_representation_factory is not None:
            assert isinstance(parent_node, AlgorithmNode) or parent_node is None
            parent_node_representation: Representation364 | None
            if parent_node is not None:
                assert isinstance(parent_node, AlgorithmNode)
                assert isinstance(parent_node.board_representation, Representation364)
                # todo remove all these ugly assert!!
                parent_node_representation = parent_node.board_representation
            else:
                parent_node_representation = None

            board_representation = (
                self.board_representation_factory.create_from_transition(
                    board=tree_node.board,
                    parent_node_representation=parent_node_representation,
                    modifications=modifications,
                )
            )

        return AlgorithmNode(
            tree_node=tree_node,
            minmax_evaluation=minmax_evaluation,
            exploration_index_data=exploration_index_data,
            board_representation=board_representation,
        )



"""
Module to create a SyzygyTable object if the path to the syzygy tables exists,
otherwise return None.
"""

import os
from typing import Any, Protocol

from chipiron.players.boardevaluators.table_base.syzygy_python import SyzygyChiTable
from chipiron.players.boardevaluators.table_base.syzygy_table import SyzygyTable
from chipiron.utils.logger import chipiron_logger
from chipiron.utils.path_variables import SYZYGY_TABLES_DIR

from .syzygy_rust import SyzygyRustTable

# Type alias for any SyzygyTable implementation
AnySyzygyTable = SyzygyTable[Any]


def create_syzygy_python() -> SyzygyChiTable | None:
    """
    Create a SyzygyTable object if the path to the syzygy tables exists,
    otherwise return None.

    Returns:
        SyzygyTable | None: The created SyzygyTable object or None if the path does not exist.
    """
    syzygy_table: SyzygyChiTable | None
    is_exist: bool = os.path.exists(SYZYGY_TABLES_DIR)

    if is_exist:
        syzygy_table = SyzygyChiTable(path_to_table=SYZYGY_TABLES_DIR)
    else:
        chipiron_logger.warning("no folder found for syzygy tables so NOT using it")
        syzygy_table = None

    return syzygy_table


def create_syzygy_rust() -> SyzygyRustTable | None:
    """
    Create a SyzygyTable object if the path to the syzygy tables exists,
    otherwise return None.

    Returns:
        SyzygyTable | None: The created SyzygyTable object or None if the path does not exist.
    """

    syzygy_table: SyzygyRustTable | None
    is_exist: bool = os.path.exists(SYZYGY_TABLES_DIR)

    if is_exist:
        syzygy_table = SyzygyRustTable(path_to_table=SYZYGY_TABLES_DIR)
    else:
        chipiron_logger.warning("no folder found for syzygy tables so NOT using it")
        syzygy_table = None

    return syzygy_table


class SyzygyProvider(Protocol):
    """Protocol for providing a Syzygy table."""

    def provide(
        self,
    ) -> AnySyzygyTable | None:
        """
        Provide a Syzygy table.
        """
        ...


class SyzygyFactory(Protocol):
    """
    Protocol for creating a Syzygy table.
    """

    def __call__(
        self,
    ) -> AnySyzygyTable | None:
        """
        Create a SyzygyTable object.
        """
        ...


def create_syzygy_factory(use_rust: bool) -> SyzygyFactory:
    """
    Create a SyzygyTable object
    """
    syzygy_factory: SyzygyFactory
    if use_rust:
        syzygy_factory = create_syzygy_rust
    else:
        syzygy_factory = create_syzygy_python

    return syzygy_factory


def create_syzygy(use_rust: bool) -> AnySyzygyTable | None:
    """
    Create a SyzygyTable object
    """
    syzygy_table: AnySyzygyTable | None
    if use_rust:
        syzygy_table = create_syzygy_rust()
    else:
        syzygy_table = create_syzygy_python()

    return syzygy_table



"""
This module defines an enumeration class for Move Selector Types.

The MoveSelectorTypes class is a subclass of the str class and the Enum class.
It represents different types of move selectors that can be used in a game.

Attributes:
    Random: Represents a random move selector.
    TreeAndValue: Represents a move selector based on tree and value.
    Stockfish: Represents a move selector using the Stockfish engine.
    CommandLineHuman: Represents a move selector for a human player using the command line.
    GuiHuman: Represents a move selector for a human player using a graphical user interface.
"""

from enum import Enum


class MoveSelectorTypes(str, Enum):
    """
    Enumeration class representing different types of move selectors.
    """

    Random = "Random"
    TreeAndValue = "TreeAndValue"
    Stockfish = "Stockfish"
    CommandLineHuman = "CommandLineHuman"
    GuiHuman = "GuiHuman"

    def is_human(self) -> bool:
        """
        Determines if the move selector type represents a human player.

        Returns:
            bool: True if the move selector type is either GuiHuman or CommandLineHuman, False otherwise.
        """
        return (
            self is MoveSelectorTypes.GuiHuman
            or self is MoveSelectorTypes.CommandLineHuman
        )



"""
This module contains the NoisyValueTreeNode class, which is a subclass of TreeNode.
"""

import chess

from chipiron.environments.chess_env.board.board_chi import BoardChi
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode
from chipiron.players.move_selector.treevalue.nodes.tree_node import TreeNode


class NoisyValueTreeNode(TreeNode):
    """
    A class representing a node in a noisy value tree.
    Inherits from TreeNode.
    """

    def __init__(
        self,
        board: BoardChi,
        half_move: int,
        id_number: int,
        parent_node: ITreeNode,
        last_move: chess.Move,
    ) -> None:
        """
        Initializes a NoisyValueTreeNode object.

        Args:
            board (BoardChi): The chess board.
            half_move (int): The half move number.
            id_number (int): The ID number of the node.
            parent_node (ITreeNode): The parent node.
            last_move (chess.Move): The last move made.

        """
        # super(TreeNode).__init__(board, half_move, id_number, parent_node, last_move)
        self.number_of_samples = 0
        self.variance = 0

    def dot_description(self) -> str:
        """
        Returns the dot description of the node.

        Returns:
            str: The dot description.

        """
        super_description: str = super().dot_description()
        return (
            super_description
            + "\n num_sampl: "
            + str(self.number_of_samples)
            + "\n variance: "
            + str(self.variance)
        )



"""
Default arguments for the script.
"""

from dataclasses import dataclass
from typing import Protocol

from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils.dataclass import IsDataclass


@dataclass
class DefaultScriptArgs:
    """Default arguments for the script."""

    base_script_args: BaseScriptArgs


class HasBaseScriptArgs(Protocol):
    """Mixin class that provides access to the base_script_args attribute."""

    base_script_args: BaseScriptArgs


class DataClassWithBaseScriptArgs(HasBaseScriptArgs, IsDataclass, Protocol):
    """Data class that includes base_script_args."""



"""
This module provides a random move selector for a chess game.

The Random class in this module implements a move selector that randomly selects a legal move from the given chess board.
It uses a random number generator to make the selection.

Example usage:
    random_selector = create_random(random_generator)
    move = random_selector.select_move(board, move_seed)

"""

import random
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal

import chipiron.environments.chess_env.board as boards
from chipiron.utils import Seed

from .move_selector import MoveRecommendation
from .move_selector_types import MoveSelectorTypes

if TYPE_CHECKING:
    from chipiron.environments.chess_env.move import moveUci
    from chipiron.environments.chess_env.move.imove import moveKey


@dataclass
class Random:
    """
    Random move selector class.

    This class implements a move selector that randomly selects a legal move from the given chess board.

    Attributes:
        type (Literal[MoveSelectorTypes.Random]): The type of move selector (for serialization).
        random_generator (random.Random): The random number generator used for making the selection.

    """

    type: Literal[MoveSelectorTypes.Random]  # for serialization
    random_generator: random.Random = field(default_factory=random.Random)

    def select_move(self, board: boards.IBoard, move_seed: Seed) -> MoveRecommendation:
        """
        Selects a random move from the given chess board.

        Args:
            board (boards.BoardChi): The chess board.
            move_seed (seed): The seed for the random number generator.

        Returns:
            MoveRecommendation: The selected move recommendation.

        """
        self.random_generator.seed(move_seed)
        random_move_key: moveKey = self.random_generator.choice(
            board.legal_moves.get_all()
        )
        random_move_uci: moveUci = board.get_uci_from_move_key(move_key=random_move_key)
        return MoveRecommendation(move=random_move_uci)


def create_random(random_generator: random.Random) -> Random:
    """
    Creates a random move selector.

    Args:
        random_generator (random.Random): The random number generator to use.

    Returns:
        Random: The created random move selector.

    """
    return Random(type=MoveSelectorTypes.Random, random_generator=random_generator)



"""Utility functions and data structures for handling chess board states and FEN strings."""

import typing
from dataclasses import dataclass, field

import chess

from chipiron.environments.chess_env.move import moveUci

fen = typing.Annotated[str, "a string representing a fen"]


@dataclass
class FenPlusMoves:
    """Represents a FEN string and its subsequent moves."""

    original_fen: fen
    subsequent_moves: list[chess.Move] = field(
        default_factory=lambda: list[chess.Move]()
    )


@dataclass
class FenPlusMoveHistory:
    """Represents a FEN string and its move history."""

    current_fen: fen
    historical_moves: list[moveUci] = field(default_factory=lambda: list[moveUci]())


@dataclass
class FenPlusHistory:
    """Represents a FEN string and its move history, along with historical board states."""

    current_fen: fen
    historical_moves: list[moveUci] = field(default_factory=lambda: list[moveUci]())
    historical_boards: list[chess._BoardState] = field(  # pyright: ignore[reportPrivateUsage]
        default_factory=lambda: list[chess._BoardState]()  # pyright: ignore[reportPrivateUsage]
    )

    def current_turn(self) -> chess.Color:
        """Returns the color of the player to move."""
        # copy of some code in the chess python library that cannot be easily extracted or called directly
        parts = self.current_fen.split()

        # Board part.
        try:
            _ = parts.pop(0)
        except IndexError:
            raise ValueError("empty fen") from None

        # Turn.
        try:
            turn_part = parts.pop(0)
        except IndexError:
            turn = chess.WHITE
        else:
            if turn_part == "w":
                turn = chess.WHITE
            elif turn_part == "b":
                turn = chess.BLACK
            else:
                raise ValueError(f"expected 'w' or 'b' for turn part of fen: {fen!r}")
        return turn


def square_rotate(square: chess.Square) -> chess.Square:
    """Rotates the square 180."""
    return square ^ 0x3F


def bitboard_rotate(bitboard: chess.Bitboard) -> chess.Bitboard:
    """Rotates the square 180."""
    return chess.flip_horizontal(bb=chess.flip_vertical(bb=bitboard))



"""
Module defining the starting position arguments for the chess board.
"""

from dataclasses import dataclass
from enum import Enum
from importlib.resources import as_file, files
from typing import Literal, Protocol

from chipiron.environments.chess_env.board.board_tools import convert_to_fen


class StartingPositionArgsType(str, Enum):
    """
    Enum class representing the type of starting position arguments.
    """

    FROMFILE = "from_file"
    FEN = "fen"


@dataclass
class StartingPositionArgs(Protocol):
    """
    Dataclass representing the base class for starting position arguments.
    """

    def get_fen(self) -> str:
        """
        Returns the FEN (Forsyth-Edwards Notation) string representing the current state of the chess board.
        Returns:
            str: The FEN string of the current board position.
        """
        ...


@dataclass
class FenStartingPositionArgs:
    """
    Dataclass representing the starting position arguments specified by FEN.
    """

    type: Literal[StartingPositionArgsType.FEN]
    fen: str

    def get_fen(self) -> str:
        """
        Returns the FEN (Forsyth-Edwards Notation) string representing the current state of the chess board.
        Returns:
            str: The FEN string of the current board position.
        """
        return self.fen


@dataclass
class FileStartingPositionArgs(StartingPositionArgs):
    """
    Dataclass representing the starting position arguments specified by a file.
    """

    type: Literal[StartingPositionArgsType.FROMFILE]
    file_name: str

    def get_fen(self) -> str:
        """
        Returns the FEN (Forsyth-Edwards Notation) string representing the current state of the chess board.
        Returns:
            str: The FEN string of the current board position.
        """
        resource = files("chipiron").joinpath("data/starting_boards/" + self.file_name)
        with as_file(resource) as real_path:
            with open(real_path, "r", encoding="utf-8") as f:
                ascii_board: str = str(f.read())
                fen: str = convert_to_fen(ascii_board)
        return fen


AllStartingPositionArgs = FenStartingPositionArgs | FileStartingPositionArgs



"""
This script defines a graphical user interface (GUI) for interacting with the chipiron program.
The GUI allows the user to select various options and perform different actions such as playing against the chipiron AI, watching a game, or visualizing a tree.

The script_gui function creates the GUI window and handles the user interactions.
The play_against_chipiron, watch_a_game, and visualize_a_tree functions are called when the corresponding buttons are clicked.

Note: This script requires the customtkinter and chipiron modules to be installed.
"""

# TODO switch to pygame
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, cast

import customtkinter as ctk
from parsley_coco import make_partial_dataclass_with_optional_paths

from chipiron import scripts
from chipiron.environments.chess_env.board.starting_position import (
    FenStartingPositionArgs,
    StartingPositionArgsType,
)
from chipiron.games.game.game_args import GameArgs
from chipiron.games.match.match_args import MatchArgs
from chipiron.games.match.match_settings_args import MatchSettingsArgs
from chipiron.games.match.match_tag import MatchConfigTag
from chipiron.players import PlayerArgs
from chipiron.players.move_selector.move_selector_types import MoveSelectorTypes
from chipiron.players.move_selector.treevalue import TreeAndValuePlayerArgs
from chipiron.players.move_selector.treevalue.progress_monitor.progress_monitor import (
    StoppingCriterionTypes,
    TreeMoveLimitArgs,
)
from chipiron.players.player_ids import PlayerConfigTag
from chipiron.scripts.one_match.one_match import MatchScriptArgs
from chipiron.scripts.script_args import BaseScriptArgs
from chipiron.utils.dataclass import IsDataclass
from chipiron.utils.logger import chipiron_logger

if TYPE_CHECKING:
    import tkinter as tk


# Move this to module level for access by callback functions
CHIPI_ALGO_OPTIONS_LIST: list[tuple[str, PlayerConfigTag]] = [
    ("RecurZipfBase3", PlayerConfigTag.RECUR_ZIPF_BASE_3),
    ("Uniform", PlayerConfigTag.UNIFORM),
    ("Sequool", PlayerConfigTag.SEQUOOL),
    ("Human Player", PlayerConfigTag.GUI_HUMAN),
]


def format_gui_args_for_display(gui_args: Any) -> str:
    """
    Format GUI arguments for human-readable display in logging.

    Args:
        gui_args: The GUI arguments object to format

    Returns:
        A formatted string representation of the arguments
    """
    if gui_args is None:
        return "None"

    args_parts: list[str] = []

    if hasattr(gui_args, "gui") and gui_args.gui:
        args_parts.append("GUI: Enabled")

    if hasattr(gui_args, "match_args") and gui_args.match_args:
        match_args = gui_args.match_args
        if hasattr(match_args, "player_one") and match_args.player_one:
            args_parts.append(
                f"Player One: {match_args.player_one.name if hasattr(match_args.player_one, 'name') else match_args.player_one}"
            )
        if hasattr(match_args, "player_two") and match_args.player_two:
            args_parts.append(
                f"Player Two: {match_args.player_two.name if hasattr(match_args.player_two, 'name') else match_args.player_two}"
            )
        if hasattr(match_args, "match_setting") and match_args.match_setting:
            args_parts.append(
                f"Match Setting: {match_args.match_setting.name if hasattr(match_args.match_setting, 'name') else match_args.match_setting}"
            )

        # Display player overwrite settings
        if (
            hasattr(match_args, "player_one_overwrite")
            and match_args.player_one_overwrite
        ):
            overwrite = match_args.player_one_overwrite
            if (
                hasattr(overwrite, "main_move_selector")
                and overwrite.main_move_selector
            ):
                selector = overwrite.main_move_selector
                if (
                    hasattr(selector, "stopping_criterion")
                    and selector.stopping_criterion
                ):
                    criterion = selector.stopping_criterion
                    if hasattr(criterion, "tree_move_limit"):
                        args_parts.append(
                            f"Player One Strength: {criterion.tree_move_limit} moves"
                        )

        if (
            hasattr(match_args, "player_two_overwrite")
            and match_args.player_two_overwrite
        ):
            overwrite = match_args.player_two_overwrite
            if (
                hasattr(overwrite, "main_move_selector")
                and overwrite.main_move_selector
            ):
                selector = overwrite.main_move_selector
                if (
                    hasattr(selector, "stopping_criterion")
                    and selector.stopping_criterion
                ):
                    criterion = selector.stopping_criterion
                    if hasattr(criterion, "tree_move_limit"):
                        args_parts.append(
                            f"Player Two Strength: {criterion.tree_move_limit} moves"
                        )

    if hasattr(gui_args, "base_script_args") and gui_args.base_script_args:
        base_args = gui_args.base_script_args
        if hasattr(base_args, "profiling") and base_args.profiling is not None:
            args_parts.append(
                f"Profiling: {'Enabled' if base_args.profiling else 'Disabled'}"
            )
        if hasattr(base_args, "seed") and base_args.seed is not None:
            args_parts.append(f"Random Seed: {base_args.seed}")

    return "\n    ".join(args_parts) if args_parts else "Default settings"


class MatchGUIStartingPosition(str, Enum):
    """The starting position for the chess match."""

    STANDARD = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"
    ENDGAME = "6k1/p7/8/8/7N/7K/2N5/8 w - - 0 1"

    def get_fen(self) -> str:
        """Get the FEN string corresponding to the starting position.

        Returns:
            str: The FEN string for the starting position.
        """
        return str(self.value)


STARTING_POSITION_OPTIONS_DICT: dict[str, MatchGUIStartingPosition] = {
    "Standard": MatchGUIStartingPosition.STANDARD,
    "End game": MatchGUIStartingPosition.ENDGAME,
}


class ScriptGUIType(str, Enum):
    """The type of script to run based on GUI selection."""

    PLAY_OR_WATCH_A_GAME = "play_or_watch_a_game"
    TREE_VISUALIZATION = "tree_visualization"


@dataclass
class ArgsChosenByUser:
    """
    The arguments chosen by the user in the GUI.
    """

    type: ScriptGUIType = ScriptGUIType.PLAY_OR_WATCH_A_GAME
    player_type_white: PlayerConfigTag = PlayerConfigTag.RECUR_ZIPF_BASE_3
    strength_white: int | None = 1
    player_type_black: PlayerConfigTag = PlayerConfigTag.RECUR_ZIPF_BASE_3
    strength_black: int | None = 1
    starting_position: MatchGUIStartingPosition = (
        MatchGUIStartingPosition.STANDARD
    )  # "Standard" or "Endgame"


def script_gui() -> tuple[scripts.ScriptType, IsDataclass | None, str]:
    """
    Creates a graphical user interface (GUI) for interacting with the chipiron program.

    Returns:
        A tuple containing the script type and the arguments for the selected action.
    """
    root: ctk.CTk = ctk.CTk()
    args_chosen_by_user: ArgsChosenByUser = ArgsChosenByUser()
    # place a label on the root window
    root.title("🐙 chess with chipirons 🐙")

    # frm = ctk.CTkFrame(root, padding=10)

    window_width: int = 900
    window_height: int = 400

    # get the screen dimension
    screen_width: int = root.winfo_screenwidth()
    screen_height: int = root.winfo_screenheight()

    # find the center point
    center_x: int = int(screen_width / 2 - window_width / 2)
    center_y: int = int(screen_height / 2 - window_height / 2)

    # set the position of the window to the center of the screen
    root.geometry(f"{window_width}x{window_height}+{center_x}+{center_y}")
    # root.iconbitmap('download.jpg')

    message = cast("tk.Widget", ctk.CTkLabel(root, text="What to do?"))
    message.grid(column=0, row=0)

    # exit button
    exit_button = ctk.CTkButton(root, text="Exit", command=lambda: root.quit())

    message = cast("tk.Widget", ctk.CTkLabel(root, text="♙ White Player: "))
    message.grid(column=0, row=2)

    message = cast("tk.Widget", ctk.CTkLabel(root, text="♟ Black Player: "))
    message.grid(column=0, row=3)

    # Create the list of options
    chipi_algo_options_list: list[tuple[str, PlayerConfigTag]] = CHIPI_ALGO_OPTIONS_LIST

    ######PLAYER WHITE

    # Variable to keep track of the option
    # selected in OptionMenu
    chipi_algo_choice_white = ctk.StringVar(
        value=args_chosen_by_user.player_type_white
    )  # Use display name as default

    # Helper to get display names and values
    algo_display_names = [name for name, _ in chipi_algo_options_list]

    # Create the option menu widget and passing the display names
    algo_menu = ctk.CTkOptionMenu(
        master=root, values=algo_display_names, variable=chipi_algo_choice_white
    )
    cast("tk.Widget", algo_menu).grid(column=3, row=2)

    strength_label = ctk.CTkLabel(root, text="  strength: ")

    # Create the list of options
    options_list = ["1", "2", "3", "4", "5"]

    # Variable to keep track of the option
    # selected in OptionMenu
    strength_value_white = ctk.StringVar(value="1")  # set initial value

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu = ctk.CTkOptionMenu(
        master=root, variable=strength_value_white, values=options_list
    )

    # Always grid the strength controls to maintain layout consistency
    cast("tk.Widget", strength_label).grid(column=4, row=2, padx=5, pady=10)
    cast("tk.Widget", strength_menu).grid(column=5, row=2, padx=10, pady=10)

    # When reading the value, map display name to PlayerConfigTag
    def get_selected_algo_tag() -> PlayerConfigTag:
        selected_name = chipi_algo_choice_white.get()
        for name, tag in chipi_algo_options_list:
            if name == selected_name:
                return tag
        return PlayerConfigTag.RECUR_ZIPF_BASE_3  # fallback

    def update_strength_menu(*_: Any) -> None:
        # Only show strength menu and label if not human
        current_tag = get_selected_algo_tag()
        if current_tag != PlayerConfigTag.GUI_HUMAN:
            strength_label.configure(text="  strength: ")
            strength_menu.configure(state="normal")
            cast("tk.Widget", strength_label).grid(column=4, row=2, padx=5, pady=10)
            cast("tk.Widget", strength_menu).grid(column=5, row=2, padx=10, pady=10)
        else:
            # Make them invisible but keep the space
            strength_label.configure(text="")
            strength_menu.grid_remove()

    chipi_algo_choice_white.trace_add("write", update_strength_menu)
    update_strength_menu()

    ######PLAYER BLACK

    # Variable to keep track of the option
    # selected in OptionMenu
    chipi_algo_choice_black = ctk.StringVar(
        value=args_chosen_by_user.player_type_black
    )  # set initial value

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    algo_menu_black = ctk.CTkOptionMenu(
        master=root, values=algo_display_names, variable=chipi_algo_choice_black
    )
    cast("tk.Widget", algo_menu_black).grid(column=3, row=3)

    strength_label_black = ctk.CTkLabel(root, text="  strength: ")

    # Variable to keep track of the option
    # selected in OptionMenu
    strength_value_black = ctk.StringVar(value="1")  # set initial value

    # Create the option menu widget and passing
    # the options_list and value_inside to it.
    strength_menu_black = ctk.CTkOptionMenu(
        master=root, variable=strength_value_black, values=options_list
    )

    # Always grid the strength controls to maintain layout consistency
    cast("tk.Widget", strength_label_black).grid(column=4, row=3, padx=5, pady=10)
    cast("tk.Widget", strength_menu_black).grid(column=5, row=3, padx=10, pady=10)

    # When reading the value, map display name to PlayerConfigTag for black player
    def get_selected_algo_tag_black() -> PlayerConfigTag:
        selected_name = chipi_algo_choice_black.get()
        for name, tag in chipi_algo_options_list:
            if name == selected_name:
                return tag
        return PlayerConfigTag.RECUR_ZIPF_BASE_3  # fallback

    def update_strength_menu_black(*_: Any) -> None:
        # Only show strength menu and label if not human
        current_tag_black = get_selected_algo_tag_black()
        if current_tag_black != PlayerConfigTag.GUI_HUMAN:
            # Make them visible
            strength_label_black.configure(text="  strength: ")
            strength_menu_black.configure(state="normal")
            cast("tk.Widget", strength_label_black).grid(
                column=4, row=3, padx=5, pady=10
            )
            cast("tk.Widget", strength_menu_black).grid(
                column=5, row=3, padx=10, pady=10
            )
        else:
            # Make them invisible but keep the space
            strength_label_black.configure(text="")
            strength_menu_black.grid_remove()

    chipi_algo_choice_black.trace_add("write", update_strength_menu_black)
    update_strength_menu_black()

    ######STARTING POSITION

    # Label for starting position
    starting_position_label = ctk.CTkLabel(root, text="Starting Position: ")
    cast("tk.Widget", starting_position_label).grid(column=0, row=4, padx=5, pady=10)

    # Variable to keep track of the starting position selection
    starting_position_choice = ctk.StringVar(value="Standard")

    # Starting position options
    starting_position_display_names = list(STARTING_POSITION_OPTIONS_DICT.keys())

    # Create the option menu widget for starting position
    starting_position_menu = ctk.CTkOptionMenu(
        master=root,
        values=starting_position_display_names,
        variable=starting_position_choice,
    )
    cast("tk.Widget", starting_position_menu).grid(column=1, row=4, padx=10, pady=10)

    # play or watch button
    play_or_watch_a_game_button: ctk.CTkButton = ctk.CTkButton(
        root,
        text="Play or Watch a game",
        command=lambda: [
            play_or_watch_a_game(
                args_chosen_by_user=args_chosen_by_user,
                chipi_algo_choice_white=chipi_algo_choice_white,
                chipi_algo_choice_black=chipi_algo_choice_black,
                strength_value_white=strength_value_white,
                strength_value_black=strength_value_black,
                starting_position_choice=starting_position_choice,
            ),
            root.destroy(),
        ],
    )

    # visualize button
    visualize_a_tree_button: ctk.CTkButton = ctk.CTkButton(
        root,
        text="Visualize a tree",
        command=lambda: [visualize_a_tree(args_chosen_by_user), root.destroy()],
    )

    cast("tk.Widget", play_or_watch_a_game_button).grid(
        row=6, column=0, padx=10, pady=10
    )
    cast("tk.Widget", visualize_a_tree_button).grid(row=8, column=0, padx=10, pady=10)
    cast("tk.Widget", exit_button).grid(row=10, column=0, padx=10, pady=10)

    cast("tk.Tk", root).mainloop()

    return generate_inputs(args_chosen_by_user=args_chosen_by_user)


def generate_inputs(
    args_chosen_by_user: ArgsChosenByUser,
) -> tuple[scripts.ScriptType, IsDataclass | None, str]:
    """Generates script type and arguments based on the user input.

    Args:
        args_chosen_by_user (ArgsChosenByUser): The arguments chosen by the user.

    Returns:
        tuple[scripts.ScriptType, IsDataclass | None, str]: The script type, arguments, and config file name.
    """
    script_type: scripts.ScriptType

    PartialOpMatchScriptArgs = make_partial_dataclass_with_optional_paths(
        cls=MatchScriptArgs
    )
    PartialOpMatchArgs = make_partial_dataclass_with_optional_paths(cls=MatchArgs)

    PartialOpMatchSettingsArgs = make_partial_dataclass_with_optional_paths(
        cls=MatchSettingsArgs
    )
    PartialOpGameArgs = make_partial_dataclass_with_optional_paths(cls=GameArgs)

    PartialOpPlayerArgs = make_partial_dataclass_with_optional_paths(cls=PlayerArgs)

    PartialOpBaseScriptArgs = make_partial_dataclass_with_optional_paths(
        cls=BaseScriptArgs
    )

    PartialOpTreeAndValuePlayerArgs = make_partial_dataclass_with_optional_paths(
        cls=TreeAndValuePlayerArgs
    )
    PartialOpTreeMoveLimitArgs = make_partial_dataclass_with_optional_paths(
        cls=TreeMoveLimitArgs
    )

    config_file_name: str
    match args_chosen_by_user.type:
        case ScriptGUIType.PLAY_OR_WATCH_A_GAME:
            gui_args = PartialOpMatchScriptArgs(
                gui=True,
                base_script_args=PartialOpBaseScriptArgs(profiling=False, seed=0),
                match_args=PartialOpMatchArgs(
                    match_setting=MatchConfigTag.DUDA,
                    match_setting_overwrite=PartialOpMatchSettingsArgs(
                        game_args=PartialOpGameArgs(
                            starting_position=FenStartingPositionArgs(
                                type=StartingPositionArgsType.FEN,
                                fen=args_chosen_by_user.starting_position,
                            )
                        )
                    ),
                ),
            )
            config_file_name = "chipiron/scripts/one_match/inputs/human_play_against_computer/exp_options.yaml"

            gui_args.match_args.player_one = PlayerConfigTag(
                args_chosen_by_user.player_type_white
            )
            gui_args.match_args.player_two = PlayerConfigTag(
                args_chosen_by_user.player_type_black
            )
            if args_chosen_by_user.player_type_white != PlayerConfigTag.GUI_HUMAN:
                assert args_chosen_by_user.strength_white is not None
                tree_move_limit_white: int = 4 * 10**args_chosen_by_user.strength_white
                gui_args.match_args.player_one_overwrite = PartialOpPlayerArgs(
                    main_move_selector=PartialOpTreeAndValuePlayerArgs(
                        type=MoveSelectorTypes.TreeAndValue,
                        stopping_criterion=PartialOpTreeMoveLimitArgs(
                            type=StoppingCriterionTypes.TREE_MOVE_LIMIT,
                            tree_move_limit=tree_move_limit_white,
                        ),
                    )
                )
            if args_chosen_by_user.player_type_black != PlayerConfigTag.GUI_HUMAN:
                assert args_chosen_by_user.strength_black is not None
                tree_move_limit_black: int = 4 * 10**args_chosen_by_user.strength_black
                gui_args.match_args.player_two_overwrite = PartialOpPlayerArgs(
                    main_move_selector=PartialOpTreeAndValuePlayerArgs(
                        type=MoveSelectorTypes.TreeAndValue,
                        stopping_criterion=PartialOpTreeMoveLimitArgs(
                            type=StoppingCriterionTypes.TREE_MOVE_LIMIT,
                            tree_move_limit=tree_move_limit_black,
                        ),
                    )
                )

            script_type = scripts.ScriptType.ONE_MATCH
        case ScriptGUIType.TREE_VISUALIZATION:
            config_file_name = "scripts/tree_visualization/inputs/base/exp_options.yaml"
            gui_args = None
            script_type = scripts.ScriptType.TREE_VISUALIZATION

    # Format arguments for human-readable display
    args_display = format_gui_args_for_display(gui_args)

    chipiron_logger.info(
        "GUI Configuration Selected:\n"
        "  Script Type: %s\n"
        "  Arguments:\n    %s\n"
        "  Config File: %s",
        script_type.name if hasattr(script_type, "name") else script_type,
        args_display,
        config_file_name,
    )
    return script_type, gui_args, config_file_name


def play_or_watch_a_game(
    args_chosen_by_user: ArgsChosenByUser,
    chipi_algo_choice_white: ctk.StringVar,
    chipi_algo_choice_black: ctk.StringVar,
    strength_value_white: ctk.StringVar,
    strength_value_black: ctk.StringVar,
    starting_position_choice: ctk.StringVar,
) -> bool:
    """
    Callback function for the "Play or Watch a game" button.
    Sets the output dictionary with the selected options for playing or watching a game.

    Args:
        output: The output dictionary to store the selected options.

    Returns:
        True.
    """
    args_chosen_by_user.type = ScriptGUIType.PLAY_OR_WATCH_A_GAME

    # Map display names to PlayerConfigTag enums
    selected_white = chipi_algo_choice_white.get()
    for name, tag in CHIPI_ALGO_OPTIONS_LIST:
        if name == selected_white:
            args_chosen_by_user.player_type_white = tag
            break
    else:
        args_chosen_by_user.player_type_white = (
            PlayerConfigTag.RECUR_ZIPF_BASE_3
        )  # fallback

    selected_black = chipi_algo_choice_black.get()
    for name, tag in CHIPI_ALGO_OPTIONS_LIST:
        if name == selected_black:
            args_chosen_by_user.player_type_black = tag
            break
    else:
        args_chosen_by_user.player_type_black = (
            PlayerConfigTag.RECUR_ZIPF_BASE_3
        )  # fallback

    args_chosen_by_user.strength_white = (
        int(strength_value_white.get())
        if args_chosen_by_user.player_type_white != PlayerConfigTag.GUI_HUMAN
        else None
    )
    args_chosen_by_user.strength_black = (
        int(strength_value_black.get())
        if args_chosen_by_user.player_type_black != PlayerConfigTag.GUI_HUMAN
        else None
    )

    # Capture starting position choice
    args_chosen_by_user.starting_position = STARTING_POSITION_OPTIONS_DICT[
        starting_position_choice.get()
    ]

    return True


def visualize_a_tree(args_chosen_by_user: ArgsChosenByUser) -> bool:
    """
    Callback function for the "Visualize a tree" button.
    Sets the output dictionary with the selected options for tree visualization.

    Args:
        output: The output dictionary to store the selected options.

    Returns:
        True.
    """
    args_chosen_by_user.type = ScriptGUIType.TREE_VISUALIZATION
    return True



"""
This module contains classes for managing update instructions in a batch.

Classes:
- UpdateInstructions: Represents update instructions for a single node.
- UpdateInstructionsBatch: Represents a batch of update instructions for multiple nodes.
"""

from dataclasses import dataclass, field
from typing import Self

from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.move_selector.treevalue.nodes import ITreeNode
from chipiron.utils.dict_of_numbered_dict_with_pointer_on_max import (
    DictOfNumberedDictWithPointerOnMax,
)

from .index_block import (
    IndexUpdateInstructionsFromOneNode,
    IndexUpdateInstructionsTowardsOneParentNode,
)
from .value_block import (
    ValueUpdateInstructionsFromOneNode,
    ValueUpdateInstructionsTowardsOneParentNode,
)


@dataclass(slots=True)
class UpdateInstructionsFromOneNode:
    """
    Represents update instructions generated from a single node.

    Attributes:
    - value_block: The value update instructions generated from a single node.
    - index_block: The index update instructions generated from a single node.
    """

    value_block: ValueUpdateInstructionsFromOneNode | None = None
    index_block: IndexUpdateInstructionsFromOneNode | None = None


@dataclass(slots=True)
class UpdateInstructionsTowardsOneParentNode:
    """
    Represents update instructions for a single node.

    Attributes:
    - value_block: The value update instructions block.
    - index_block: The index update instructions block.
    """

    value_updates_toward_one_parent_node: (
        ValueUpdateInstructionsTowardsOneParentNode | None
    ) = None
    index_updates_toward_one_parent_node: (
        IndexUpdateInstructionsTowardsOneParentNode | None
    ) = None

    def add_update_from_a_child_node(
        self,
        update_from_a_child_node: UpdateInstructionsFromOneNode,
        move_from_parent_to_child: moveKey,
    ) -> None:
        assert self.value_updates_toward_one_parent_node is not None
        assert update_from_a_child_node.value_block is not None
        self.value_updates_toward_one_parent_node.add_update_from_one_child_node(
            move_from_parent_to_child=move_from_parent_to_child,
            update_from_one_child_node=update_from_a_child_node.value_block,
        )

        if self.index_updates_toward_one_parent_node is None:
            assert update_from_a_child_node.index_block is None
        else:
            if update_from_a_child_node.index_block is not None:
                self.index_updates_toward_one_parent_node.add_update_from_one_child_node(
                    move_from_parent_to_child=move_from_parent_to_child,
                    update_from_one_child_node=update_from_a_child_node.index_block,
                )

    def add_updates_towards_one_parent_node(self, another_update: Self) -> None:
        assert self.value_updates_toward_one_parent_node is not None
        assert another_update.value_updates_toward_one_parent_node is not None
        self.value_updates_toward_one_parent_node.add_update_toward_one_parent_node(
            another_update.value_updates_toward_one_parent_node
        )

        if self.index_updates_toward_one_parent_node is None:
            assert another_update.index_updates_toward_one_parent_node is None
        else:
            if another_update.index_updates_toward_one_parent_node is not None:
                self.index_updates_toward_one_parent_node.add_update_toward_one_parent_node(
                    another_update.index_updates_toward_one_parent_node
                )

    def print_info(self) -> None:
        """
        Prints information about the update instructions.
        """
        print("printing info of update instructions")
        assert (
            self.index_updates_toward_one_parent_node is not None
            and self.value_updates_toward_one_parent_node is not None
        )
        self.value_updates_toward_one_parent_node.print_info()
        self.index_updates_toward_one_parent_node.print_info()

    def empty(self) -> bool:
        """
        Checks if the update instructions are empty.

        Returns:
        - True if the update instructions are empty, False otherwise.
        """
        assert self.value_updates_toward_one_parent_node is not None
        return self.value_updates_toward_one_parent_node.empty() and (
            self.index_updates_toward_one_parent_node is None
            or self.index_updates_toward_one_parent_node.empty()
        )


@dataclass
class UpdateInstructionsTowardsMultipleNodes:
    """Represents update instructions towards multiple parent nodes."""

    # mapping from nodes to the update instructions that are intended to them for consideration (performing the updates)
    one_node_instructions: DictOfNumberedDictWithPointerOnMax[
        ITreeNode, UpdateInstructionsTowardsOneParentNode
    ] = field(
        default_factory=lambda: DictOfNumberedDictWithPointerOnMax[
            ITreeNode, UpdateInstructionsTowardsOneParentNode
        ]()
    )

    def add_update_from_one_child_node(
        self,
        update_from_child_node: UpdateInstructionsFromOneNode,
        parent_node: ITreeNode,
        move_from_parent: moveKey,
    ) -> None:
        if parent_node not in self.one_node_instructions:
            # build the UpdateInstructionsTowardsOneParentNode
            assert update_from_child_node.value_block is not None
            value_updates_toward_one_parent_node: (
                ValueUpdateInstructionsTowardsOneParentNode
            )
            value_updates_toward_one_parent_node = (
                ValueUpdateInstructionsTowardsOneParentNode(
                    moves_with_updated_value=(
                        {move_from_parent}
                        if update_from_child_node.value_block.new_value_for_node
                        else set()
                    ),
                    moves_with_updated_over=(
                        {move_from_parent}
                        if update_from_child_node.value_block.is_node_newly_over
                        else set()
                    ),
                    moves_with_updated_best_move=(
                        {move_from_parent}
                        if update_from_child_node.value_block.new_best_move_for_node
                        else set()
                    ),
                )
            )
            index_updates_toward_one_parent_node: (
                IndexUpdateInstructionsTowardsOneParentNode | None
            )
            if update_from_child_node.index_block is not None:
                index_updates_toward_one_parent_node = (
                    IndexUpdateInstructionsTowardsOneParentNode(
                        moves_with_updated_index=(
                            {move_from_parent}
                            if update_from_child_node.index_block.updated_index
                            else set()
                        ),
                    )
                )
            else:
                index_updates_toward_one_parent_node = None
            update_instructions_towards_parent: UpdateInstructionsTowardsOneParentNode
            update_instructions_towards_parent = UpdateInstructionsTowardsOneParentNode(
                value_updates_toward_one_parent_node=value_updates_toward_one_parent_node,
                index_updates_toward_one_parent_node=index_updates_toward_one_parent_node,
            )
            self.one_node_instructions[parent_node] = update_instructions_towards_parent
        else:
            # update the UpdateInstructionsTowardsOneParentNode
            self.one_node_instructions[parent_node].add_update_from_a_child_node(
                update_from_a_child_node=update_from_child_node,
                move_from_parent_to_child=move_from_parent,
            )

    def add_updates_towards_one_parent_node(
        self,
        update_from_child_node: UpdateInstructionsTowardsOneParentNode,
        parent_node: ITreeNode,
    ) -> None:
        if parent_node in self.one_node_instructions:
            self.one_node_instructions[parent_node].add_updates_towards_one_parent_node(
                another_update=update_from_child_node
            )
        else:
            self.one_node_instructions[parent_node] = update_from_child_node

    def pop_item(self) -> tuple[ITreeNode, UpdateInstructionsTowardsOneParentNode]:
        return self.one_node_instructions.popitem()

    def __bool__(self) -> bool:
        """
        Checks if the data structure is non-empty.

        Returns:
            bool: True if the data structure is non-empty, False otherwise.
        """
        return bool(self.one_node_instructions)



"""
Module in charge of managing the game. It is the main class that will be used to play a game.
"""

import os
import queue
from dataclasses import asdict
from typing import TYPE_CHECKING

import chess
import yaml

import chipiron.players as players_m
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.environments.chess_env.move_factory import MoveFactory
from chipiron.games.game.game_playing_status import PlayingStatus
from chipiron.players.boardevaluators.board_evaluator import IGameBoardEvaluator
from chipiron.players.boardevaluators.table_base.factory import AnySyzygyTable
from chipiron.utils import path
from chipiron.utils.communication.gui_messages import (
    BackMessage,
    GameStatusMessage,
    PlayerProgressMessage,
)
from chipiron.utils.communication.player_game_messages import MoveMessage
from chipiron.utils.dataclass import IsDataclass, custom_asdict_factory
from chipiron.utils.logger import chipiron_logger

from .final_game_result import FinalGameResult, GameReport
from .game import ObservableGame
from .game_args import GameArgs
from .progress_collector import PlayerProgressCollectorP

if TYPE_CHECKING:
    from chipiron.environments import HalfMove
    from chipiron.environments.chess_env.board.iboard import IBoard
    from chipiron.environments.chess_env.move import moveUci


class GameManager:
    """
    Object in charge of playing one game
    """

    # The game object that is managed
    game: ObservableGame

    # A SyzygyTable
    syzygy: AnySyzygyTable | None

    # Evaluators that just evaluates the boards but are not players (just spectators) for display info of who is winning
    # according to them
    display_board_evaluator: IGameBoardEvaluator

    # folder to log results
    output_folder_path: path | None

    # args of the Game
    args: GameArgs

    # Dictionary mapping colors to player names?
    player_color_to_id: dict[chess.Color, str]

    # A Queue for receiving messages from other process or functions such as players or Gui
    main_thread_mailbox: queue.Queue[IsDataclass]

    # The list of players
    players: list[players_m.PlayerProcess | players_m.GamePlayer]

    # A move factory
    move_factory: MoveFactory

    # an object for collecting how advances each player is in its thinking/computation of the moves
    progress_collector: PlayerProgressCollectorP

    def __init__(
        self,
        game: ObservableGame,
        syzygy: AnySyzygyTable | None,
        display_board_evaluator: IGameBoardEvaluator,
        output_folder_path: path | None,
        args: GameArgs,
        player_color_to_id: dict[chess.Color, str],
        main_thread_mailbox: queue.Queue[IsDataclass],
        players: list[players_m.PlayerProcess | players_m.GamePlayer],
        move_factory: MoveFactory,
        progress_collector: PlayerProgressCollectorP,
    ) -> None:
        """
        Constructor for the GameManager Class. If the args, and players are not given a value it is set to None,
        waiting for the set methods to be called. This is done like this so that the players can be changed
        (often swapped) easily.

        Args:
            game (ObservableGame): The observable game object.
            syzygy (AnySyzygyTable | None): The syzygy table object or None.
            display_board_evaluator (IGameBoardEvaluator): The board evaluator to display an independent evaluation.
            output_folder_path (path | None): The output folder path or None.
            args (GameArgs): The game arguments.
            player_color_to_id (dict[chess.Color, str]): The dictionary mapping player color to player ID.
            main_thread_mailbox (queue.Queue[IsDataclass]): The main thread mailbox.
            players (list[players_m.PlayerProcess | players_m.GamePlayer]): The list of players.

        Returns:
            None
        """
        self.game = game
        self.syzygy = syzygy
        self.path_to_store_result = (
            os.path.join(output_folder_path, "games")
            if output_folder_path is not None
            else None
        )
        self.display_board_evaluator = display_board_evaluator
        self.args = args
        self.player_color_to_id = player_color_to_id
        self.main_thread_mailbox = main_thread_mailbox
        self.players = players
        self.move_factory = move_factory
        self.progress_collector = progress_collector

    def external_eval(self) -> tuple[float | None, float]:
        """Evaluates the game board using the display board evaluator.

        Returns:
            tuple[float, float]: A tuple containing the evaluation scores.
        """
        return self.display_board_evaluator.evaluate(self.game.board)

    def play_one_move(self, move: moveKey) -> None:
        """Play one move in the game.

        Args:
            move (chess.Move): The move to be played.
        """
        self.game.play_move(move)
        if self.syzygy is not None and self.syzygy.fast_in_table(self.game.board):
            chipiron_logger.info(
                "Theoretically finished with value for white: %s",
                self.syzygy.string_result(self.game.board),
            )

    def rewind_one_move(self) -> None:
        """
        Rewinds the game by one move.

        This method rewinds the game by one move, undoing the last move made.
        If the game has a Syzygy tablebase loaded and the current board position is in the tablebase,
        it prints the theoretically finished value for white.

        Returns:
            None
        """
        self.game.rewind_one_move()
        if self.syzygy is not None and self.syzygy.fast_in_table(self.game.board):
            chipiron_logger.info(
                "Theoretically finished with value for white: %s",
                self.syzygy.string_result(self.game.board),
            )

    def play_one_game(self) -> GameReport:
        """
        Play one game.

        Returns:
            GameReport: The report of the game.
        """

        color_names = ["Black", "White"]

        # sending the current board to the gui
        self.game.notify_display()

        # sending the current board to the player and asking for a move
        if self.game.is_play():
            self.game.query_move_from_players()

        while True:
            board = self.game.board
            half_move: HalfMove = board.ply()
            chipiron_logger.info(
                "Half Move: %s playing status %s",
                half_move,
                self.game.playing_status.status,
            )
            color_to_move: chess.Color = board.turn
            color_of_player_to_move_str = color_names[color_to_move]
            chipiron_logger.info(
                "%s (%s) to play now...",
                color_of_player_to_move_str,
                self.player_color_to_id[color_to_move],
            )

            # waiting for a message
            mail = self.main_thread_mailbox.get()
            self.processing_mail(mail)

            board = self.game.board
            if board.is_game_over() or not self.game_continue_conditions():
                if board.is_game_over():
                    chipiron_logger.info("The game is over")
                if not self.game_continue_conditions():
                    chipiron_logger.info("Game continuation not met")
                break
            chipiron_logger.info("Not game over at %s", board)

        self.tell_results()
        self.terminate_processes()
        chipiron_logger.info("End play_one_game")

        game_results: FinalGameResult = self.simple_results()

        game_report: GameReport = GameReport(
            final_game_result=game_results,
            move_history=[move for move in self.game.move_history],
            fen_history=self.game.fen_history,
        )
        return game_report

    def processing_mail(self, message: IsDataclass) -> None:
        """
        Process the incoming mail message.

        Args:
            message (IsDataclass): The incoming mail message.

        Returns:
            None
        """

        board: IBoard = self.game.board

        match message:
            case MoveMessage():
                chipiron_logger.info(
                    "=====================MOVE MESSAGE RECEIVED============"
                )
                move_message: MoveMessage = message
                # play the move
                move_uci: moveUci = move_message.move

                chipiron_logger.info(
                    "Game Manager: Receiving the move uci %s %s %s",
                    move_uci,
                    self.game.playing_status,
                    board.fen,
                )
                if (
                    move_message.corresponding_board == board.fen
                    and self.game.playing_status.is_play()
                    and message.player_name == self.player_color_to_id[board.turn]
                ):
                    board.legal_moves.get_all()  # make sure the board has generated the legal moves

                    move_key: moveKey = board.get_move_key_from_uci(move_uci=move_uci)

                    chipiron_logger.info(
                        "Game Manager: Play a move %s at %s %s",
                        move_uci,
                        board,
                        self.game.board.fen,
                    )
                    # move: IMove = self.move_factory(move_uci=move_uci, board=board)
                    self.play_one_move(move_key)
                    chipiron_logger.info(
                        "Game Manager: Now board is  %s", self.game.board
                    )

                    eval_sto, eval_chi = self.external_eval()
                    chipiron_logger.info(
                        "Stockfish evaluation: %s and chipiron eval %s",
                        eval_sto,
                        eval_chi,
                    )
                    # Print the board
                    chipiron_logger.info(board.print_chess_board())

                    # sending the current board to the player  and asking for a move
                    if self.game.is_play():
                        self.game.query_move_from_players()

                else:
                    chipiron_logger.info(
                        "the move is rejected because one of the following is false \n"
                        " move_message.corresponding_board == board.fen%s \n"
                        "self.game.playing_status.is_play() %s\n"
                        "message.player_name == self.player_color_to_id[board.turn] %s",
                        move_message.corresponding_board == board.fen,
                        self.game.playing_status.is_play(),
                        message.player_name == self.player_color_to_id[board.turn],
                    )
                    chipiron_logger.info(
                        "%s,%s",
                        message.player_name,
                        self.player_color_to_id[board.turn],
                    )
                if message.evaluation is not None:
                    self.display_board_evaluator.add_evaluation(
                        player_color=message.color_to_play,
                        evaluation=message.evaluation,
                    )
            case PlayerProgressMessage():
                player_progress_message: PlayerProgressMessage = message
                if player_progress_message.player_color == chess.WHITE:
                    self.progress_collector.progress_white(
                        value=player_progress_message.progress_percent
                    )
                if player_progress_message.player_color == chess.BLACK:
                    self.progress_collector.progress_black(
                        value=player_progress_message.progress_percent
                    )
            case GameStatusMessage():
                game_status_message: GameStatusMessage = message
                # update game status
                if game_status_message.status == PlayingStatus.PLAY:
                    self.game.set_play_status()
                    self.game.query_move_from_players()

                if game_status_message.status == PlayingStatus.PAUSE:
                    self.game.set_pause_status()
            case BackMessage():
                self.game.set_pause_status()
                self.rewind_one_move()

            case other:
                raise ValueError(
                    f"type of message received is not recognized {other} in file {__name__}"
                )

    def game_continue_conditions(self) -> bool:
        """
        Checks the conditions for continuing the game.

        Returns:
            bool: True if the game should continue, False otherwise.
        """
        half_move: HalfMove = self.game.board.ply()
        continue_bool: bool = True
        if (
            self.args.max_half_moves is not None
            and half_move >= self.args.max_half_moves
        ):
            continue_bool = False

        return continue_bool

    def print_to_file(self, game_report: GameReport, idx: int = 0) -> None:
        """
        Print the moves of the game to a yaml file and a more human-readable text file.

        Args:
            game_report(GameReport): a game report to be printed
            idx (int): The index to include in the file name (default is 0).

        Returns:
            None
        """
        # todo probably the txt file should be a valid PGN file : https://en.wikipedia.org/wiki/Portable_Game_Notation
        if self.path_to_store_result is not None:
            path_file: path = (
                f"{self.path_to_store_result}_{idx}_W:{self.player_color_to_id[chess.WHITE]}"
                f"-vs-B:{self.player_color_to_id[chess.BLACK]}"
            )
            path_file_obj = f"{path_file}_game_report.yaml"
            path_file_txt = f"{path_file}.txt"
            with open(path_file_txt, "a", encoding="utf-8") as the_fileText:
                move_1 = None
                for counter, move in enumerate(self.game.move_history):
                    if counter % 2 == 0:
                        move_1 = move
                    else:
                        if move_1 is not None:
                            the_fileText.write(str(move_1) + " " + str(move) + "\n")
            with open(path_file_obj, "w", encoding="utf-8") as file:
                yaml.dump(
                    asdict(game_report, dict_factory=custom_asdict_factory),
                    file,
                    default_flow_style=False,
                )

    def tell_results(self) -> None:
        """
        Prints the results of the game based on the current state of the board.

        The method checks various conditions on the board and prints the corresponding result.
        It also checks for specific conditions like syzygy, fivefold repetition, seventy-five moves,
        insufficient material, stalemate, and checkmate.

        Returns:
            None
        """
        board = self.game.board
        if self.syzygy is not None and self.syzygy.fast_in_table(board):
            chipiron_logger.info(
                "Syzygy: Theoretical value for white %s",
                self.syzygy.string_result(board),
            )
        board.tell_result()

    def simple_results(self) -> FinalGameResult:
        """
        Determines the final result of the game based on the current state of the board.

        Returns:
            FinalGameResult: The final result of the game.
        """
        board = self.game.board

        res: FinalGameResult | None = None
        result: str = board.result(claim_draw=True)
        if result == "*":
            if self.syzygy is None or not self.syzygy.fast_in_table(board):
                # useful when a game is stopped
                # before the end, for instance for debugging and profiling
                res = FinalGameResult.DRAW  # arbitrary meaningless choice
                # raise ValueError(f'Problem with figuring our game results in {__name__}')
            else:
                raise ValueError(
                    "this case is not coded atm think of what is the right thing to do here!"
                )
        else:
            match result:
                case "1/2-1/2":
                    res = FinalGameResult.DRAW
                case "0-1":
                    res = FinalGameResult.WIN_FOR_BLACK
                case "1-0":
                    res = FinalGameResult.WIN_FOR_WHITE
                case other:
                    raise ValueError(
                        f"unexpected result value {other} in game manager/simple_results"
                    )

        assert isinstance(res, FinalGameResult)
        return res

    def terminate_processes(self) -> None:
        """
        Terminates all player processes and stops the associated threads.

        This method iterates over the list of players and terminates any player processes found.
        If a player process is found, it is terminated and the associated thread is stopped.

        Note:
        - This method assumes that the `players` attribute is a list of `PlayerProcess` or `GamePlayer` objects.

        Returns:
        None
        """
        player: players_m.PlayerProcess | players_m.GamePlayer
        for player in self.players:
            if isinstance(player, players_m.PlayerProcess):
                player.terminate()
                chipiron_logger.info("Stopping the thread")
                chipiron_logger.info("Thread stopped")



"""
This module contains the implementation of the NodeEvaluator class, which is responsible for evaluating the value of
 nodes in a tree-based move selector.

The NodeEvaluator class wraps a board evaluator and a syzygy table to provide more complex evaluations of chess
 positions. It handles queries for evaluating nodes and manages obvious over events.

Classes:
- NodeEvaluator: Wrapping node evaluator with syzygy and obvious over event.

Enums:
- NodeEvaluatorTypes: Types of node evaluators.

Constants:
- DISCOUNT: Discount factor used in the evaluation.

Functions:
- None

"""

from enum import Enum
from typing import TYPE_CHECKING

from chipiron.players.boardevaluators.master_board_evaluator import MasterBoardEvaluator
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)

if TYPE_CHECKING:
    from chipiron.players.boardevaluators.over_event import OverEvent

DISCOUNT = 0.99999999  # lokks like at the moment the use is to break ties in the evaluation (not sure if needed or helpful now)


class NodeEvaluatorTypes(str, Enum):
    """
    Enum class representing different types of node evaluators.
    """

    NEURAL_NETWORK = "neural_network"


class EvaluationQueries:
    """
    A class that represents evaluation queries for algorithm nodes.

    Attributes:
        over_nodes (list[AlgorithmNode]): A list of algorithm nodes that are considered "over".
        not_over_nodes (list[AlgorithmNode]): A list of algorithm nodes that are not considered "over".
    """

    over_nodes: list[AlgorithmNode]
    not_over_nodes: list[AlgorithmNode]

    def __init__(self) -> None:
        """
        Initializes a new instance of the NodeEvaluator class.
        """
        self.over_nodes = []
        self.not_over_nodes = []

    def clear_queries(self) -> None:
        """
        Clears the evaluation queries by resetting the over_nodes and not_over_nodes lists.
        """
        self.over_nodes = []
        self.not_over_nodes = []


class NodeEvaluator:
    """
    The NodeEvaluator class is responsible for evaluating the value of nodes in a tree structure.
    It uses a board evaluator and a syzygy evaluator to calculate the value of the nodes.
    """

    master_board_evaluator: MasterBoardEvaluator

    def __init__(
        self,
        master_board_evaluator: MasterBoardEvaluator,
    ) -> None:
        """
        Initializes a NodeEvaluator object.

        Args:
            board_evaluator (MasterBoardEvaluator): The board evaluator used to evaluate the chess board.
        """
        self.master_board_evaluator = master_board_evaluator

    def check_obvious_over_events(self, node: AlgorithmNode) -> None:
        """
        Updates the node.over object if the game is obviously over.
        """
        over_event: OverEvent | None
        evaluation: float | None
        over_event, evaluation = self.master_board_evaluator.check_obvious_over_events(
            node.board
        )
        if over_event is not None:
            node.minmax_evaluation.over_event.becomes_over(
                how_over=over_event.how_over,
                who_is_winner=over_event.who_is_winner,
                termination=over_event.termination,
            )
            assert evaluation is not None, (
                "Evaluation should not be None for over nodes"
            )
            node.minmax_evaluation.set_evaluation(evaluation=evaluation)

    def evaluate_all_queried_nodes(self, evaluation_queries: EvaluationQueries) -> None:
        """
        Evaluates all the queried nodes.
        """
        # node_over: AlgorithmNode
        # for node_over in evaluation_queries.over_nodes:
        # assert isinstance(node_over, AlgorithmNode)
        #    self.evaluate_over(node_over)

        if evaluation_queries.not_over_nodes:
            self.evaluate_all_not_over(evaluation_queries.not_over_nodes)

        evaluation_queries.clear_queries()

    def add_evaluation_query(
        self, node: AlgorithmNode, evaluation_queries: EvaluationQueries
    ) -> None:
        """
        Adds an evaluation query for a node.
        """
        assert node.minmax_evaluation.value_white_evaluator is None
        self.check_obvious_over_events(node)
        if node.is_over():
            evaluation_queries.over_nodes.append(node)
        else:
            evaluation_queries.not_over_nodes.append(node)

    def evaluate_all_not_over(self, not_over_nodes: list[AlgorithmNode]) -> None:
        """
        Evaluates all the nodes that are not over.
        """

        node_not_over: AlgorithmNode
        for node_not_over in not_over_nodes:
            evaluation = self.master_board_evaluator.value_white(
                board=node_not_over.board
            )
            processed_evaluation = self.process_evalution_not_over(
                evaluation=evaluation, node=node_not_over
            )
            # assert isinstance(node_not_over, AlgorithmNode)
            node_not_over.minmax_evaluation.set_evaluation(processed_evaluation)

    def process_evalution_not_over(
        self, evaluation: float, node: AlgorithmNode
    ) -> float:
        """
        Processes the evaluation for a node that is not over.
        """
        processed_evaluation = (1 / DISCOUNT) ** node.half_move * evaluation
        return processed_evaluation



"""
Module where we define the Stockfish Board Evaluator
"""

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal

import chess.engine

import chipiron.environments.chess_env.board as boards
from chipiron.environments.chess_env.board.factory import create_board_chi
from chipiron.environments.chess_env.board.utils import FenPlusHistory
from chipiron.players.boardevaluators.board_evaluator_type import BoardEvalTypes
from chipiron.utils.logger import chipiron_logger

if TYPE_CHECKING:
    from chipiron.environments.chess_env.board.board_chi import BoardChi


@dataclass
class StockfishBoardEvalArgs:
    """Represents the arguments for the Stockfish board evaluator.

    Attributes:
        depth (int): The depth of the search algorithm.
        time_limit (float): The time limit for the search algorithm.
    """

    # for deserialization
    type: Literal[BoardEvalTypes.STOCKFISH_BOARD_EVAL] = field(init=False)

    depth: int
    time_limit: float

    def __post_init__(self) -> None:
        """
        Post-initialization method for the dataclass.
        Automatically sets the 'type' attribute to BoardEvalTypes.STOCKFISH_BOARD_EVAL
        to avoid manual assignment and to facilitate discrimination during deserialization.
        """

        # to avoid having to set the type mannually in the code as it is obvious and only used for discrimanating in deserialization
        object.__setattr__(self, "type", BoardEvalTypes.STOCKFISH_BOARD_EVAL)


class StockfishBoardEvaluator:
    """
    A board evaluator powered by stockfish
    """

    engine: chess.engine.SimpleEngine | None
    args: StockfishBoardEvalArgs

    def __init__(self, args: StockfishBoardEvalArgs) -> None:
        """
        Initializes a StockfishBoardEvaluator object.

        Args:
            args (StockfishBoardEvalArgs): The arguments for the StockfishBoardEvaluator.

        Returns:
            None
        """
        self.args = args
        self.engine = None

    def value_white(self, board: boards.IBoard) -> float:
        """
        Computes the value of the board for the white player.

        Args:
            board (boards.BoardChi): The board object representing the current state of the game.

        Returns:
            float: The value of the board for the white player.
        """
        try:
            if self.engine is None:
                # Try multiple possible Stockfish paths
                stockfish_paths = [
                    "/usr/games/stockfish",
                    "stockfish",
                    r"stockfish/stockfish/stockfish-ubuntu-x86-64-avx2",
                ]

                for path in stockfish_paths:
                    try:
                        self.engine = chess.engine.SimpleEngine.popen_uci(path)
                        break
                    except (FileNotFoundError, OSError):
                        continue

                if self.engine is None:
                    chipiron_logger.warning(
                        "Stockfish binary not found in any of the expected locations."
                    )
                    return 0.0  # Fallback if no Stockfish found

            # Use the board's chess_board directly to avoid state inconsistencies
            # transform the board
            board_chi: BoardChi = create_board_chi(
                fen_with_history=FenPlusHistory(
                    current_fen=board.fen,
                    # historical_moves=board.move_history_stack,
                    # historical_boards=board.board_history_stack,  # type: ignore
                    # note that we do not give here historical_boards, hope this does not create but related to 3 fold repetition computation
                )
            )

            info = self.engine.analyse(
                board_chi.chess_board, chess.engine.Limit(time=0.01)
            )

            score = info.get("score")

            if score is None:
                return 0.0

            # Convert score to float
            white_score = score.white()

            if score.is_mate():
                # For mate scores, return a large value based on mate distance
                mate_moves = white_score.mate()
                if mate_moves is not None:
                    if mate_moves > 0:
                        return 1000.0 - mate_moves  # White wins
                    else:
                        return -1000.0 - mate_moves  # Black wins
                return 0.0
            else:
                # Convert centipawns to pawns (divide by 100)
                cp_value = getattr(white_score, "cp", 0)
                return float(cp_value) / 100.0

        except (chess.engine.EngineError, FileNotFoundError, OSError):
            return 0.0  # Fallback on any error



"""
This module contains the definition of the NNPytorchTrainer class, which is responsible for training and testing a neural network model using PyTorch.
"""

import typing
from typing import Callable

import torch
from torch.utils.data import DataLoader

from chipiron.players.boardevaluators.datasets.datasets import SupervisedData
from chipiron.utils.chi_nn import ChiNN
from chipiron.utils.logger import chipiron_logger


def compute_loss(
    net: ChiNN,
    criterion: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    input_layer: torch.Tensor,
    target_value: torch.Tensor,
) -> torch.Tensor:
    prediction: torch.Tensor = net(input_layer)
    loss: torch.Tensor = criterion(prediction, target_value)
    return loss


def check_model_device(model: ChiNN) -> str | torch.device | int:
    # Check the device of the first parameter

    first_param_device = next(model.parameters()).device

    return first_param_device


def compute_test_error_on_dataset(
    net: ChiNN,
    criterion: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    data_test: DataLoader[SupervisedData],
    number_of_tests: int = 100,
) -> float:
    """Compute the test error on the given dataset.

    Args:
        net (ChiNN): The neural network model to be tested.
        criterion (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): The loss function used for testing.
        data_test (DataLoader[SupervisedData]): The test dataset.
        number_of_tests (int, optional): The number of tests to run. Defaults to 100.

    Returns:
        float: The test error on the given dataset.
    """
    sum_loss_test = 0.0
    count_test = 0
    loss_test: torch.Tensor
    device = check_model_device(net)
    for _ in range(number_of_tests):
        sample = next(iter(data_test))
        input_layer, target_value = (
            sample.get_input_layer().to(device),
            sample.get_target_value().to(device),
        )

        loss_test = compute_loss(
            net=net,
            criterion=criterion,
            input_layer=input_layer,
            target_value=target_value,
        )
        sum_loss_test += float(loss_test)
        count_test += 1
    chipiron_logger.info("test error %f", float(sum_loss_test / float(count_test)))
    test_error: float = float(sum_loss_test / float(count_test))
    return test_error


class NNPytorchTrainer:
    """
    A class that trains a neural network model using PyTorch.

    Args:
        net (ChiNN): The neural network model to be trained.
        optimizer (torch.optim.Optimizer): The optimizer used for updating the model's parameters.
        scheduler (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler.

    Attributes:
        net (ChiNN): The neural network model to be trained.
        criterion (torch.nn.L1Loss): The loss function used for training.
        optimizer (torch.optim.Optimizer): The optimizer used for updating the model's parameters.
        scheduler (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler.

    Methods:
        train(input_layer, target_value): Trains the neural network model using the provided input and target values.
        test(input_layer, target_value): Tests the neural network model using the provided input and target values.
        train_next_boards(input_layer, next_input_layer): Trains the neural network model using the provided input and next input layers.
    """

    def __init__(
        self,
        net: ChiNN,
        optimizer: torch.optim.Optimizer,
        scheduler: torch.optim.lr_scheduler.LRScheduler,
    ) -> None:
        """
        Initializes a new instance of the NNPytorchTrainer class.

        Args:
            net (ChiNN): The neural network model to be trained.
            optimizer (torch.optim.Optimizer): The optimizer used for updating the model's parameters.
            scheduler (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler.

        Returns:
            None
        """
        self.net = net
        self.criterion = torch.nn.L1Loss()
        self.optimizer = optimizer
        self.scheduler = scheduler

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.net.to(self.device)
        chipiron_logger.info(f"Model put to device {self.device}")

    @typing.no_type_check
    def train(
        self, input_layer: torch.Tensor, target_value: torch.Tensor
    ) -> torch.Tensor:
        """
        Trains the neural network model using the provided input and target values.

        Args:
            input_layer (torch.Tensor): The input data.
            target_value (torch.Tensor): The target values.

        Returns:
            torch.Tensor: The loss value.
        """
        self.net.train()

        self.optimizer.zero_grad()
        input_layer, target_value = (
            input_layer.to(self.device),
            target_value.to(self.device),
        )

        loss: torch.Tensor = compute_loss(
            net=self.net,
            criterion=self.criterion,
            input_layer=input_layer,
            target_value=target_value,
        )
        loss.backward()

        self.optimizer.step()
        return loss

    def test(
        self, input_layer: torch.Tensor, target_value: torch.Tensor
    ) -> torch.Tensor:
        """
        Tests the neural network model using the provided input and target values.

        Args:
            input_layer (torch.Tensor): The input data.
            target_value (torch.Tensor): The target values.

        Returns:
            torch.Tensor: The loss value.
        """
        self.net.eval()
        input_layer, target_value = (
            input_layer.to(self.device),
            target_value.to(self.device),
        )
        loss: torch.Tensor = compute_loss(
            net=self.net,
            criterion=self.criterion,
            input_layer=input_layer,
            target_value=target_value,
        )
        self.net.train()
        return loss

    def train_next_boards(
        self, input_layer: torch.Tensor, next_input_layer: torch.Tensor
    ) -> None:
        """
        Trains the neural network model using the provided input and next input layers.

        Args:
            input_layer (torch.Tensor): The input data.
            next_input_layer (torch.Tensor): The next input data.

        Returns:
            None
        """
        self.net.eval()
        target_value = -self.net(next_input_layer)

        self.net.train()
        self.optimizer.zero_grad()
        prediction_with_player_to_move_as_white = self.net(input_layer)

        loss = self.criterion(prediction_with_player_to_move_as_white, target_value)
        loss.backward()
        self.optimizer.step()

    def compute_test_error_on_dataset(
        self, data_test: DataLoader[SupervisedData]
    ) -> float:
        """
        Computes the test error of the neural network model.

        This method iterates over a test dataset and calculates the average loss
        for a given number of iterations. The test error is then computed as the
        average loss divided by the number of iterations.

        Returns:
            None
        """

        self.net.eval()
        test_error: float = compute_test_error_on_dataset(
            net=self.net, criterion=self.criterion, data_test=data_test
        )
        self.net.train()
        return test_error



"""
This module contains classes and functions related to opening instructions in a chess game.
"""

import random
from dataclasses import dataclass
from enum import Enum
from typing import Any, ItemsView, Iterator, Self, ValuesView

import chipiron.players.move_selector.treevalue.nodes as nodes
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.move_selector.treevalue.nodes.utils import (
    a_move_key_sequence_from_root,
    a_move_uci_sequence_from_root,
)


@dataclass(slots=True)
class OpeningInstruction:
    """
    Represents an opening instruction for a specific node in the game tree.
    """

    node_to_open: nodes.ITreeNode
    move_to_play: moveKey

    def print_info(self) -> None:
        """
        Prints information about the opening instruction.
        """
        print(
            f"OpeningInstruction: node_to_open {self.node_to_open.id} at hm {self.node_to_open.half_move} {self.node_to_open.board.fen}| "
            f"a path from root to node_to_open is {a_move_key_sequence_from_root(self.node_to_open)} {a_move_uci_sequence_from_root(self.node_to_open)}| "
            f"self.move_to_play {self.move_to_play} {self.node_to_open.board.get_uci_from_move_key(self.move_to_play)}"
        )


class OpeningInstructions:
    # todo do we need a dict? why not a set? verify

    """
    Represents a collection of opening instructions.
    """

    batch: dict[Any, OpeningInstruction]

    def __init__(self, dictionary: dict[Any, OpeningInstruction] | None = None) -> None:
        """
        Initializes the OpeningInstructions object.

        Args:
            dictionary: A dictionary of opening instructions (optional).
        """
        # here i use a dictionary because they are insertion ordered until there is an ordered set in python
        # order is important because some method give a batch where the last element in the batch are prioritary
        self.batch = {}

        if dictionary is not None:
            for key in dictionary:
                self[key] = dictionary[key]

    def __setitem__(self, key: Any, value: OpeningInstruction) -> None:
        """
        Sets an opening instruction in the collection.

        Args:
            key: The key for the opening instruction.
            value: The opening instruction.
        """
        # key is supposed to be a tuple with (node_to_open,  move_to_play)
        assert len(key) == 2
        self.batch[key] = value

    def __getitem__(self, key: Any) -> OpeningInstruction:
        """
        Retrieves an opening instruction from the collection.

        Args:
            key: The key for the opening instruction.

        Returns:
            The opening instruction.
        """
        # assert(0==1)
        return self.batch[key]

    def __iter__(self) -> Iterator[Any]:
        """
        Returns an iterator over the keys of the opening instructions.

        Returns:
            An iterator over the keys.
        """
        return iter(self.batch)

    def __bool__(self) -> bool:
        """
        Checks if the collection is non-empty.

        Returns:
            True if the collection is non-empty, False otherwise.
        """
        return bool(self.batch)

    def merge(self, another_opening_instructions_batch: Self) -> None:
        """
        Merges another batch of opening instructions into the current collection.

        Args:
            another_opening_instructions_batch: Another OpeningInstructions object.
        """
        for (
            opening_instruction_key,
            opening_instruction,
        ) in another_opening_instructions_batch.items():
            if opening_instruction_key not in self.batch:
                self.batch[opening_instruction_key] = opening_instruction

    def pop_items(self, how_many: int, popped: Self) -> None:
        """
        Pops a specified number of opening instructions from the collection.

        Args:
            how_many: The number of opening instructions to pop.
            popped: An OpeningInstructions object to store the popped instructions.
        """
        how_many = min(how_many, len(self.batch))
        for _ in range(how_many):
            key, value = self.batch.popitem()
            popped[key] = value

    def values(self) -> ValuesView[OpeningInstruction]:
        """
        Returns a view of the values in the collection.

        Returns:
            A view of the values.
        """
        return self.batch.values()

    def items(self) -> ItemsView[Any, OpeningInstruction]:
        """
        Returns a view of the items (key-value pairs) in the collection.

        Returns:
            A view of the items.
        """
        return self.batch.items()

    def print_info(self) -> None:
        """
        Prints information about the opening instructions in the collection.
        """
        print("OpeningInstructionsBatch: batch contains", len(self.batch), "elements:")
        for _key, opening_instructions in self.batch.items():
            opening_instructions.print_info()

    def __len__(self) -> int:
        """
        Returns the number of opening instructions in the collection.

        Returns:
            The number of opening instructions.
        """
        return len(self.batch)


def create_instructions_to_open_all_moves(
    moves_to_play: list[moveKey], node_to_open: nodes.ITreeNode
) -> OpeningInstructions:
    """
    Creates opening instructions for all possible moves to play from a given node.

    Args:
        moves_to_play: A list of chess moves.
        node_to_open: The node to open.

    Returns:
        An OpeningInstructions object containing the opening instructions.
    """
    opening_instructions_batch = OpeningInstructions()

    for move_to_play in moves_to_play:
        # at the moment it looks redundant keys are almost the same as values but its clean
        # the keys are here for fast and redundant proof insertion
        # and the values are here for clean data processing
        opening_instructions_batch[(node_to_open.id, move_to_play)] = (
            OpeningInstruction(node_to_open=node_to_open, move_to_play=move_to_play)
        )
    #  node_to_open.non_opened_legal_moves.add(move_to_play)
    return opening_instructions_batch


class OpeningType(Enum):
    """
    Represents the type of opening to use.
    """

    ALL_CHILDREN = "all_children"


class OpeningInstructor:
    """
    Represents an opening instructor that provides opening instructions based on a specific opening type.
    """

    def __init__(
        self, opening_type: OpeningType, random_generator: random.Random
    ) -> None:
        """
        Initializes the OpeningInstructor object.

        Args:
            opening_type: The type of opening to use.
            random_generator: A random number generator.
        """
        self.opening_type = opening_type
        self.random_generator = random_generator

    def all_moves_to_open(self, node_to_open: nodes.ITreeNode) -> list[moveKey]:
        """
        Returns a list of all possible moves to open from a given node.

        Args:
            node_to_open: The node to open.

        Returns:
            A list of chess moves.
        """
        if self.opening_type == OpeningType.ALL_CHILDREN:
            node_to_open.all_legal_moves_generated = True
            moves_to_play: list[moveKey] = node_to_open.legal_moves.get_all()

            # this shuffling add randomness to the playing style
            # (it stills depends on the random seed, but if random seed varies then the behavior will be more random)
            # DEACTIVATED ATM BECAUSE I DO NOT UNDERSTAND or FORGOT THE USE CASE: MAYBE DEAD SINCE SEED SYSTEM CHANGED
            # self.random_generator.shuffle(moves_to_play)

        else:
            raise Exception("Hello-la")
        return moves_to_play



"""
This module provides a factory function for creating the main move selector based on the given arguments.
"""

import queue
import random
from typing import TypeAlias

from chipiron.players.boardevaluators.table_base.factory import AnySyzygyTable
from chipiron.utils.logger import chipiron_logger

from ...utils.dataclass import IsDataclass
from . import human, move_selector, stockfish, treevalue
from .random import Random, create_random

AllMoveSelectorArgs: TypeAlias = (
    treevalue.TreeAndValuePlayerArgs
    | human.CommandLineHumanPlayerArgs
    | human.GuiHumanPlayerArgs
    | Random
    | stockfish.StockfishPlayer
)


def create_main_move_selector(
    move_selector_instance_or_args: AllMoveSelectorArgs,
    syzygy: AnySyzygyTable | None,
    random_generator: random.Random,
    queue_progress_player: queue.Queue[IsDataclass] | None,
) -> move_selector.MoveSelector:
    """
    Create the main move selector based on the given arguments.

    Args:
        move_selector_instance_or_args (AllMoveSelectorArgs): The arguments or instance of the move selector.
        syzygy (SyzygyTable | None): The syzygy table.
        random_generator (random.Random): The random number generator.

    Returns:
        move_selector.MoveSelector: The main move selector.

    Raises:
        ValueErr    or: If the given move selector instance or arguments are invalid.

    """
    main_move_selector: move_selector.MoveSelector
    chipiron_logger.debug("Create main move selector")

    match move_selector_instance_or_args:
        case Random():
            main_move_selector = create_random(random_generator=random_generator)
        case treevalue.TreeAndValuePlayerArgs():
            tree_args: treevalue.TreeAndValuePlayerArgs = move_selector_instance_or_args
            main_move_selector = treevalue.create_tree_and_value_builders(
                args=tree_args,
                syzygy=syzygy,
                random_generator=random_generator,
                queue_progress_player=queue_progress_player,
            )
        case stockfish.StockfishPlayer():
            main_move_selector = move_selector_instance_or_args
        case human.CommandLineHumanPlayerArgs():
            main_move_selector = human.CommandLineHumanMoveSelector()
        case other:
            raise ValueError(
                f"player creator: can not find {other} of type {type(other)}"
            )
    return main_move_selector



from chipiron.scripts.league.runtheleague import RunTheLeagueScript

from .base_tree_exploration.base_tree_exploration import BaseTreeExplorationScript
from .iscript import IScript
from .learn_from_scratch_value_and_fixed_boards.learn_from_scratch_value_and_fixed_boards import (
    LearnNNFromScratchScript,
)
from .learn_nn_supervised.learn_nn_from_supervised_datasets import LearnNNScript
from .one_match.one_match import OneMatchScript
from .replay_game.replay_game import ReplayGameScript
from .script_type import ScriptType
from .tree_visualization.tree_visualizer import VisualizeTreeScript

# script_type_to_script_class_name maps Script Type to the class name of the script to instantiate
script_type_to_script_class_name: dict[ScriptType, type[IScript]] = {
    ScriptType.ONE_MATCH: OneMatchScript,
    ScriptType.TREE_VISUALIZATION: VisualizeTreeScript,
    ScriptType.LEARN_NN: LearnNNScript,
    ScriptType.REPLAY_MATCH: ReplayGameScript,
    ScriptType.LEAGUE: RunTheLeagueScript,
    ScriptType.BASE_TREE_EXPLORATION: BaseTreeExplorationScript,
    ScriptType.LEARN_NN_FROM_SCRATCH: LearnNNFromScratchScript,
}


def get_script_type_from_script_class_name(script_type: ScriptType) -> type[IScript]:
    """
    Retrieves the script class name based on the given script type.

    Args:
        script_type (ScriptType): The script type.

    Returns:
        Any: The script class name.

    Raises:
        Exception: If the script type is not found.
    """
    if script_type in script_type_to_script_class_name:
        script_class_name: type[IScript] = script_type_to_script_class_name[script_type]
        return script_class_name
    raise KeyError(f"Cannot find the script type {script_type} in file {__name__}")



"""
Module that contains the classes for the exploration data of a tree node.
"""

import typing
from dataclasses import dataclass, field

from chipiron.utils.small_tools import Interval

if typing.TYPE_CHECKING:
    from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode
    from chipiron.players.move_selector.treevalue.nodes.tree_node import TreeNode


@dataclass
class NodeExplorationData:
    """
    Represents the exploration data for a tree node.

    Attributes:
        tree_node (TreeNode): The tree node associated with the exploration data.
        index (float | None): The index value associated with the node. Defaults to None.

    Methods:
        dot_description(): Returns a string representation of the exploration data for dot visualization.
    """

    tree_node: "TreeNode[ITreeNode]"
    index: float | None = None

    def dot_description(self) -> str:
        """
        Returns a string representation of the dot description for the index.

        Returns:
            str: The dot description of the index.
        """
        return f"index:{self.index}"


@dataclass
class RecurZipfQuoolExplorationData(NodeExplorationData):
    """
    Represents the exploration data for a tree node with recursive zipf-quool factor.

    Attributes:
        zipf_factored_proba (float | None): The probability associated with the node, factored by zipf-quool factor.
            Defaults to None.

    Methods:
        dot_description(): Returns a string representation of the exploration data for dot visualization.
    """

    # the 'proba' associated by recursively multiplying 1/rank of the node with the max zipf_factor of the parents
    zipf_factored_proba: float | None = None

    def dot_description(self) -> str:
        """
        Returns a string representation of the index and zipf_factored_proba values.

        Returns:
            str: A string representation of the index and zipf_factored_proba values.
        """
        return f"index:{self.index} zipf_factored_proba:{self.zipf_factored_proba}"


@dataclass
class MinMaxPathValue(NodeExplorationData):
    """
    Represents the exploration data for a tree node with minimum and maximum path values.

    Attributes:
        min_path_value (float | None): The minimum path value associated with the node. Defaults to None.
        max_path_value (float | None): The maximum path value associated with the node. Defaults to None.

    Methods:
        dot_description(): Returns a string representation of the exploration data for dot visualization.
    """

    min_path_value: float | None = None
    max_path_value: float | None = None

    def dot_description(self) -> str:
        return f"min_path_value: {self.min_path_value}, max_path_value: {self.max_path_value}"


@dataclass
class IntervalExplo(NodeExplorationData):
    """
    Represents the exploration data for a tree node with an interval.

    Attributes:
        interval (Interval | None): The interval associated with the node. Defaults to None.

    Methods:
        dot_description(): Returns a string representation of the exploration data for dot visualization.
    """

    interval: Interval | None = field(default_factory=Interval)

    def dot_description(self) -> str:
        """
        Returns a string representation of the interval values.

        If the interval is None, returns 'None'.
        Otherwise, returns a string in the format 'min_interval_value: {min_value}, max_interval_value: {max_value}'.

        Returns:
            str: A string representation of the interval values.
        """
        if self.interval is None:
            return "None"
        else:
            return f"min_interval_value: {self.interval.min_value}, max_interval_value: {self.interval.max_value}"


@dataclass
class MaxDepthDescendants(NodeExplorationData):
    """
    Represents the exploration data for a tree node with maximum depth of descendants.
    """

    max_depth_descendants: int = 0

    def update_from_child(self, child_max_depth_descendants: int) -> bool:
        """
        Updates the max_depth_descendants value based on the child's max_depth_descendants.

        Args:
            child_max_depth_descendants (int): The max_depth_descendants value of the child node.

        Returns:
            bool: True if the max_depth_descendants value has changed, False otherwise.
        """
        previous_index = self.max_depth_descendants
        new_index: int = max(
            self.max_depth_descendants, child_max_depth_descendants + 1
        )
        self.max_depth_descendants = new_index
        has_index_changed: bool = new_index != previous_index

        return has_index_changed

    def dot_description(self) -> str:
        """
        Returns a string representation of the dot description for the node indices.

        Returns:
            str: The dot description for the node indices.
        """
        return f"max_depth_descendants: {self.max_depth_descendants}"



from chipiron.players.boardevaluators.neural_networks.factory import (
    NeuralNetModelsAndArchitecture,
)
from chipiron.scripts.evaluate_models.evaluate_models import evaluate_models

test_models_to_evaluate_: list[NeuralNetModelsAndArchitecture] = [
    NeuralNetModelsAndArchitecture.build_from_folder_path(
        folder_path="chipiron/scripts/learn_nn_supervised/board_evaluators_common_training_data/nn_pytorch/prelu_no_bug"
    )
]


def test_evaluate_model() -> None:
    open(
        "chipiron/scripts/evaluate_models/tests/test_evaluation_report.yaml", "w"
    ).close()
    evaluate_models(
        models_to_evaluate=test_models_to_evaluate_,
        evaluation_report_file="chipiron/scripts/evaluate_models/tests/test_evaluation_report.yaml",
        dataset_file_name="chipiron/scripts/learn_nn_supervised/tests/small_dataset.pi",
    )


if __name__ == "__main__":
    test_evaluate_model()



"""
This module provides a uniform selection strategy for nodes in a tree.

The Uniform selection strategy selects nodes uniformly at random from the available options.

Example usage:
    from .uniform import Uniform

    uniform_selector = Uniform()
    selected_node = uniform_selector.select_node(nodes)

"""

from .uniform import Uniform

__all__ = ["Uniform"]



"""
This module contains classes related to tree expansion in a chess game.
"""

import typing
from dataclasses import dataclass, field

import chipiron.environments.chess_env.board as board_mod
import chipiron.players.move_selector.treevalue.nodes as node
from chipiron.environments.chess_env.move.imove import moveKey


@dataclass(slots=True)
class TreeExpansion:
    """
    Represents an expansion of a tree in a chess game.

    Attributes:
        child_node (node.ITreeNode): The child node created during the expansion.
        parent_node (node.ITreeNode | None): The parent node of the child node. None if it's the root node.
        board_modifications (board_mod.BoardModification | None): The modifications made to the chess board during the expansion.
        creation_child_node (bool): Indicates whether the child node was created during the expansion.
        move (chess.Move): the move from parent to child node.
    """

    child_node: node.ITreeNode
    parent_node: node.ITreeNode | None
    board_modifications: board_mod.BoardModificationP | None
    creation_child_node: bool
    move: moveKey | None

    def __repr__(self) -> str:
        return (
            f"child_node{self.child_node.id} | "
            f"parent_node{self.parent_node.id if self.parent_node is not None else None} | "
            f"creation_child_node{self.creation_child_node}"
        )


@dataclass(slots=True)
class TreeExpansions:
    """
    Represents a collection of tree expansions in a chess game.

    Attributes:
        expansions_with_node_creation (List[TreeExpansion]): List of expansions where child nodes were created.
        expansions_without_node_creation (List[TreeExpansion]): List of expansions where child nodes were not created.
    """

    expansions_with_node_creation: list[TreeExpansion] = field(
        default_factory=lambda: list[TreeExpansion]()
    )
    expansions_without_node_creation: list[TreeExpansion] = field(
        default_factory=lambda: list[TreeExpansion]()
    )

    def __iter__(self) -> typing.Iterator[TreeExpansion]:
        return iter(
            self.expansions_with_node_creation + self.expansions_without_node_creation
        )

    def add(self, tree_expansion: TreeExpansion) -> None:
        """
        Adds a tree expansion to the collection.

        Args:
            tree_expansion (TreeExpansion): The tree expansion to add.
        """
        if tree_expansion.creation_child_node:
            self.add_creation(tree_expansion=tree_expansion)
        else:
            self.add_connection(tree_expansion=tree_expansion)

    def add_creation(self, tree_expansion: TreeExpansion) -> None:
        """
        Adds a tree expansion with a created child node to the collection.

        Args:
            tree_expansion (TreeExpansion): The tree expansion to add.
        """
        self.expansions_with_node_creation.append(tree_expansion)

    def add_connection(self, tree_expansion: TreeExpansion) -> None:
        """
        Adds a tree expansion without a created child node to the collection.

        Args:
            tree_expansion (TreeExpansion): The tree expansion to add.
        """
        self.expansions_without_node_creation.append(tree_expansion)

    def __str__(self) -> str:
        return (
            f"expansions_with_node_creation {self.expansions_with_node_creation} \n"
            f"expansions_without_node_creation{self.expansions_without_node_creation}"
        )



"""
Module to create and save neural network trainers and their parameters
"""

import os.path
import pickle
from dataclasses import asdict, dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Iterable, List, Union, cast, no_type_check

import torch
import torch.optim as optim
import yaml

from chipiron.learningprocesses.nn_trainer.nn_trainer import NNPytorchTrainer
from chipiron.players.boardevaluators.board_evaluation.board_evaluation import (
    PointOfView,
)
from chipiron.players.boardevaluators.neural_networks.factory import (
    get_nn_architecture_file_path_from,
    get_nn_param_file_path_from,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.ModelInputRepresentationType import (
    ModelInputRepresentationType,
)
from chipiron.players.boardevaluators.neural_networks.models.multi_layer_perceptron import (
    MultiLayerPerceptronArgs,
)
from chipiron.players.boardevaluators.neural_networks.neural_net_architecture_args import (
    NeuralNetArchitectureArgs,
)
from chipiron.players.boardevaluators.neural_networks.NNModelType import (
    ActivationFunctionType,
    NNModelType,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.model_output_type import (
    ModelOutputType,
)
from chipiron.utils import path
from chipiron.utils.chi_nn import ChiNN
from chipiron.utils.dataclass import custom_asdict_factory
from chipiron.utils.logger import chipiron_logger
from chipiron.utils.small_tools import mkdir_if_not_existing

SerializableType = Union[
    str, int, float, bool, None, Dict[str, Any], List[Any], set[Any], frozenset[Any]
]


@dataclass
class NNTrainerArgs:
    """
    Arguments for the NNTrainer class.

    Attributes:
        reuse_existing_trainer (bool): Whether to reuse an existing trainer.
        starting_lr (float): The starting learning rate.
        momentum_op (float): The momentum value.
        scheduler_step_size (int): The step size for the scheduler.
        scheduler_gamma (float): The gamma value for the scheduler.
        saving_intermediate_copy (bool): Whether to save intermediate copies.
    """

    neural_network_architecture_args: NeuralNetArchitectureArgs = field(
        default_factory=lambda: NeuralNetArchitectureArgs(
            model_type_args=MultiLayerPerceptronArgs(
                type=NNModelType.MULTI_LAYER_PERCEPTRON,
                number_neurons_per_layer=[5, 1],
                list_of_activation_functions=[
                    ActivationFunctionType.TANGENT_HYPERBOLIC
                ],
            ),
            model_output_type=ModelOutputType(point_of_view=PointOfView.PLAYER_TO_MOVE),
            model_input_representation_type=ModelInputRepresentationType.PIECE_DIFFERENCE,
        )
    )
    nn_parameters_file_if_reusing_existing_one: path | None = None
    specific_saving_folder: path | None = None
    reuse_existing_model: bool = False
    reuse_existing_trainer: bool = False
    starting_lr: float = 0.1
    momentum_op: float = 0.9
    scheduler_step_size: int = 1
    scheduler_gamma: float = 0.5
    saving_intermediate_copy: bool = True

    batch_size_train: int = 32
    batch_size_test: int = 10
    saving_interval: int = 1000
    saving_intermediate_copy_interval: int = 10000
    min_interval_lr_change: int = 1000000
    min_lr: float = 0.001

    epochs_number: int = 100

    def __post_init__(self) -> None:
        if self.reuse_existing_model:
            if self.nn_parameters_file_if_reusing_existing_one is None:
                raise Exception(
                    "Problem because you are asking for a reuse of existing model without specifying a"
                    f" param file as we have: reuse_existing_model {self.reuse_existing_model}"
                    f" nn_param_file_if_not_reusing_existing_one {self.nn_parameters_file_if_reusing_existing_one}"
                )


def get_optimizer_file_path_from(folder_path: path) -> str:
    """
    Returns the file path for the optimizer file in the given folder path.

    Args:
        folder_path (str): The path to the folder containing the optimizer file.

    Returns:
        str: The file path for the optimizer file.

    """
    file_path: str = os.path.join(folder_path, "optimizer.pi")
    return file_path


def get_scheduler_file_path_from(folder_path: path) -> str:
    """Get the file path for the scheduler file in the given folder path.

    Args:
        folder_path (str): The path of the folder containing the scheduler file.

    Returns:
        str: The file path of the scheduler file.
    """
    file_path: str = os.path.join(folder_path, "scheduler.pi")
    return file_path


def get_folder_training_copies_path_from(folder_path: path) -> str:
    """
    Returns the path to the 'training_copies' folder within the given folder path.

    Parameters:
        folder_path (str): The path to the folder.

    Returns:
        str: The path to the 'training_copies' folder.
    """
    return os.path.join(folder_path, "training_copies")


def create_nn_trainer(
    args: NNTrainerArgs, nn: ChiNN, saving_folder: path
) -> NNPytorchTrainer:
    """
    Creates an instance of NNPytorchTrainer based on the provided arguments and neural network.

    Args:
        args (NNTrainerArgs): The arguments for the NNTrainer.
        nn (ChiNN): The neural network to be trained.

    Returns:
        NNPytorchTrainer: An instance of NNPytorchTrainer.

    """

    optimizer: torch.optim.Optimizer
    scheduler: torch.optim.lr_scheduler.LRScheduler
    if args.reuse_existing_trainer:
        file_optimizer_path = get_optimizer_file_path_from(folder_path=saving_folder)
        with open(file_optimizer_path, "rb") as file_optimizer:
            optimizer = pickle.load(file_optimizer)

        file_scheduler_path = get_scheduler_file_path_from(folder_path=saving_folder)
        with open(file_scheduler_path, "rb") as file_scheduler:
            scheduler = pickle.load(file_scheduler)

    else:
        optimizer = optim.SGD(
            nn.parameters(),
            lr=args.starting_lr,
            momentum=args.momentum_op,
            weight_decay=0.000,
        )
        scheduler = optim.lr_scheduler.StepLR(
            optimizer, step_size=args.scheduler_step_size, gamma=args.scheduler_gamma
        )

    if args.saving_intermediate_copy:
        folder_path_training_copies = get_folder_training_copies_path_from(
            saving_folder
        )
        mkdir_if_not_existing(folder_path_training_copies)

    return NNPytorchTrainer(net=nn, optimizer=optimizer, scheduler=scheduler)


@no_type_check
def serialize_for_yaml(obj: Any) -> SerializableType:
    """
    Recursively converts Enums and other non-serializable objects
    into basic types for safe YAML dumping.
    """
    # Handle None and primitives
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return cast("SerializableType", obj)

    # Handle Enums
    if isinstance(obj, Enum):
        return serialize_for_yaml(obj.value)

    # Handle dictionaries
    if isinstance(obj, dict):
        dict_obj = cast("Dict[Any, Any]", obj)
        dict_result: Dict[str, SerializableType] = {}
        for key, value in dict_obj.items():
            str_key: str = str(key)
            serialized_value: SerializableType = serialize_for_yaml(value)
            dict_result[str_key] = serialized_value
        return dict_result

    # Handle lists and tuples
    if isinstance(obj, (list, tuple)):
        iterable_obj: Iterable[Any] = cast("Iterable[Any]", obj)
        list_result: list[SerializableType] = [
            serialize_for_yaml(item) for item in iterable_obj
        ]
        return list_result

    # Handle frozensets
    if isinstance(obj, frozenset):
        frozenset_obj = cast("frozenset[Any]", obj)
        frozenset_result: frozenset[SerializableType] = frozenset(
            serialize_for_yaml(item) for item in frozenset_obj
        )
        return frozenset_result

    # Handle objects with __dict__
    if hasattr(obj, "__dict__"):
        return serialize_for_yaml(vars(obj))

    # Fallback: convert to string
    return str(obj)


def safe_nn_architecture_save(
    nn_architecture_args: NeuralNetArchitectureArgs, nn_param_folder_name: path
) -> None:
    """
    Save the architecture of a neural network to a file.
    Args:
        nn_architecture_args (NeuralNetArchitectureArgs): The architecture arguments of the neural network.
        nn_param_folder_name (path): The folder path where the architecture file will be saved.
    """
    path_to_param_file = get_nn_architecture_file_path_from(nn_param_folder_name)
    try:
        chipiron_logger.info("saving architecture to file: %s", path_to_param_file)
        with open(path_to_param_file, "w", encoding="utf-8") as file_architecture:
            yaml.dump(
                asdict(
                    nn_architecture_args,
                    dict_factory=custom_asdict_factory,
                ),
                file_architecture,
                default_flow_style=False,
            )
    except KeyboardInterrupt:
        exit(-1)


def safe_nn_param_save(
    nn: ChiNN,
    nn_param_folder_name: path,
    file_name: str | None = None,
    training_copy: bool = False,
) -> None:
    """
    Save the parameters of a neural network to a file.

    Args:
        nn (ChiNN): The neural network to save.
        training_copy (bool, optional): Whether to save a training copy of the parameters. Defaults to False.
    """
    folder_path = nn_param_folder_name
    folder_path_training_copies = get_folder_training_copies_path_from(folder_path)

    nn_file_path_pt: str
    file_name_yaml: str
    nn_file_path_pt, file_name_yaml = get_nn_param_file_path_from(
        folder_path=folder_path, file_name=file_name
    )
    path_to_param_file: path
    if training_copy:
        now = datetime.now()  # current date and time
        path_to_param_file = os.path.join(
            folder_path_training_copies, now.strftime("%A-%m-%d-%Y--%H:%M:%S:%f")
        )
    else:
        path_to_param_file = nn_file_path_pt
    try:
        chipiron_logger.info(f"saving to file: {path_to_param_file}")
        with open(path_to_param_file, "wb") as fileNNW:
            torch.save(nn.state_dict(), fileNNW)
            nn.log_readable_model_weights_to_file(file_path=file_name_yaml)
        with open(path_to_param_file + "_save", "wb") as fileNNW:
            torch.save(nn.state_dict(), fileNNW)
    except KeyboardInterrupt:
        with open(path_to_param_file + "_save", "wb") as fileNNW:
            torch.save(nn.state_dict(), fileNNW)
        exit(-1)


def safe_nn_trainer_save(nn_trainer: NNPytorchTrainer, nn_folder_path: path) -> None:
    """
    Safely saves the optimizer and scheduler of the given NNPytorchTrainer object to files.

    Args:
        nn_trainer (NNPytorchTrainer): The NNPytorchTrainer object containing the optimizer and scheduler to be saved.

    Returns:
        None
    """
    file_optimizer_path = get_optimizer_file_path_from(nn_folder_path)
    file_scheduler_path = get_scheduler_file_path_from(nn_folder_path)
    try:
        with open(file_optimizer_path, "wb") as file_optimizer:
            pickle.dump(nn_trainer.optimizer, file_optimizer)
        with open(file_scheduler_path, "wb") as file_scheduler:
            pickle.dump(nn_trainer.scheduler, file_scheduler)
        with open(str(file_optimizer_path) + "_save", "wb") as file_optimizer:
            pickle.dump(nn_trainer.optimizer, file_optimizer)
        with open(file_scheduler_path + "_save", "wb") as file_scheduler:
            pickle.dump(nn_trainer.scheduler, file_scheduler)
    except KeyboardInterrupt:
        with open(file_optimizer_path + "_save", "wb") as file_optimizer:
            pickle.dump(nn_trainer.optimizer, file_optimizer)
        with open(file_scheduler_path + "_save", "wb") as file_scheduler:
            pickle.dump(nn_trainer.scheduler, file_scheduler)
        exit(-1)



"""
This module defines the TreeNode class, which represents a node in a tree structure for a chess game.
"""

from dataclasses import dataclass, field

import chess

import chipiron.environments.chess_env.board as boards
from chipiron.environments.chess_env.board.iboard import LegalMoveKeyGeneratorP
from chipiron.environments.chess_env.move.imove import moveKey

from .itree_node import ITreeNode

# todo replace the any with a defaut value in ITReenode when availble in python; 3.13?


@dataclass(slots=True)
class TreeNode[ChildrenType: ITreeNode = ITreeNode]:
    r"""
    The TreeNode class stores information about a specific board position, including the board representation,
    the player to move, the half-move count, and the parent-child relationships with other nodes.

    Attributes:
        id\_ (int): The number to identify this node for easier debugging.
        half_move\_ (int): The number of half-moves since the start of the game to reach the board position.
        board\_ (boards.BoardChi): The board representation of the node.
        parent_nodes\_ (set[ITreeNode]): The set of parent nodes to this node.
        all_legal_moves_generated (bool): A boolean indicating whether all moves have been generated.
        non_opened_legal_moves (set[chess.Move]): The set of non-opened legal moves.
        moves_children\_ (dict[chess.Move, ITreeNode | None]): The dictionary mapping moves to child nodes.
        fast_rep (str): The fast representation of the board.
        player_to_move\_ (chess.Color): The color of the player that has to move in the board.

    Methods:
        __post_init__(): Initializes the TreeNode object after it has been created.
        id(): Returns the id of the node.
        player_to_move(): Returns the color of the player to move.
        board(): Returns the board representation.
        half_move(): Returns the number of half-moves.
        moves_children(): Returns the dictionary mapping moves to child nodes.
        parent_nodes(): Returns the set of parent nodes.
        is_root_node(): Checks if the node is a root node.
        legal_moves(): Returns the legal moves of the board.
        add_parent(new_parent_node: ITreeNode): Adds a parent node to the current node.
        is_over(): Checks if the game is over.
        print_moves_children(): Prints the moves-children links of the node.
        test(): Performs a test on the node.
        dot_description(): Returns the dot description of the node.
        test_all_legal_moves_generated(): Tests if all legal moves have been generated.
        get_descendants(): Returns a dictionary of descendants of the node.
    """

    # id is a number to identify this node for easier debug
    id_: int

    # number of half-moves since the start of the game to get to the board position in self.board
    half_move_: int

    # the node represents a board position. we also store the fast representation of the board.
    board_: boards.IBoard

    # the set of parent nodes to this node. Note that a node can have multiple parents!
    parent_nodes_: dict[ITreeNode[ChildrenType], moveKey]

    # all_legal_moves_generated  is a boolean saying whether all moves have been generated.
    # If true the moves are either opened in which case the corresponding opened node is stored in
    # the dictionary self.moves_children, otherwise it is stored in self.non_opened_legal_moves
    all_legal_moves_generated: bool = False
    non_opened_legal_moves: set[moveKey] = field(default_factory=lambda: set[moveKey]())

    # dictionary mapping moves to children nodes. Node is set to None if not created
    moves_children_: dict[moveKey, ChildrenType | None] = field(
        default_factory=lambda: dict[moveKey, ChildrenType | None]()
    )

    # the color of the player that has to move in the board
    player_to_move_: chess.Color = field(default_factory=chess.Color)

    def __post_init__(self) -> None:
        """
        Performs post-initialization tasks for the TreeNode class.

        This method is automatically called after the object is initialized.
        It sets the `fast_rep` attribute based on the board's fast representation
        and assigns the current player to `player_to_move_`.

        Parameters:
            None

        Returns:
            None
        """
        if self.board_:
            self.player_to_move_: chess.Color = self.board_.turn

    @property
    def fast_rep(self) -> boards.boardKey:
        """Returns the fast representation of the board.

        Returns:
            boards.boardKey: The fast representation of the board.
        """
        return self.board_.fast_representation

    @property
    def id(self) -> int:
        """
        Returns the ID of the tree node.

        Returns:
            int: The ID of the tree node.
        """
        return self.id_

    @property
    def player_to_move(self) -> chess.Color:
        """
        Returns the color of the player who is to make the next move.

        Returns:
            chess.Color: The color of the player who is to make the next move.
        """
        return self.player_to_move_

    @property
    def board(self) -> boards.IBoard:
        """
        Returns the board associated with this tree node.

        Returns:
            boards.BoardChi: The board associated with this tree node.
        """
        return self.board_

    @property
    def half_move(self) -> int:
        """
        Returns the number of half moves made in the game.

        Returns:
            int: The number of half moves made.
        """
        return self.half_move_

    @property
    def moves_children(self) -> dict[moveKey, ChildrenType | None]:
        """
        Returns a bidirectional dictionary containing the children nodes of the current tree node,
        along with the corresponding chess moves that lead to each child node.

        Returns:
            dict[chess.Move, ITreeNode | None]: A bidirectional dictionary mapping chess moves to
            the corresponding child nodes. If a move does not have a corresponding child node, it is
            mapped to None.
        """
        return self.moves_children_

    @property
    def parent_nodes(self) -> dict[ITreeNode[ChildrenType], moveKey]:
        """
        Returns the dictionary of parent nodes of the current tree node with associated move.

        :return: A dictionary of parent nodes of the current tree node with associated move.
        """
        return self.parent_nodes_

    def is_root_node(self) -> bool:
        """
        Check if the current node is a root node.

        Returns:
            bool: True if the node is a root node, False otherwise.
        """
        return not self.parent_nodes

    # understand what is best :
    # 1. this function returns a object of type chess.LegalMoveGenerator
    # 2. of type set of moves
    @property
    def legal_moves(self) -> LegalMoveKeyGeneratorP:
        """
        Returns a generator that yields the legal moves for the current board state.

        Returns:
            chess.LegalMoveGenerator: A generator that yields the legal moves.
        """
        return self.board_.legal_moves

    def add_parent(
        self, move: moveKey, new_parent_node: ITreeNode[ChildrenType]
    ) -> None:
        """
        Adds a new parent node to the current node.

        Args:
            move (chess.Move): the move that led to the node from the new_parent_node
            new_parent_node (ITreeNode): The new parent node to be added.

        Raises:
            AssertionError: If the new parent node is already in the parent nodes set.

        Returns:
            None
        """
        # debug
        assert (
            new_parent_node not in self.parent_nodes
        )  # there cannot be two ways to link the same child-parent
        self.parent_nodes[new_parent_node] = move

    def is_over(self) -> bool:
        """
        Checks if the game is over.

        Returns:
            bool: True if the game is over, False otherwise.
        """
        return self.board.is_game_over()

    def print_moves_children(self) -> None:
        """
        Prints the moves-children link of the node.

        This method prints the moves-children link of the node, showing the move and the ID of the child node.
        If a child node is None, it will be displayed as 'None'.

        Returns:
            None
        """
        print(
            "here are the ",
            len(self.moves_children_),
            " moves-children link of node",
            self.id,
            ": ",
            end=" ",
        )
        for move, child in self.moves_children_.items():
            if child is None:
                print(move, child, end=" ")
            else:
                print(move, child.id, end=" ")
        print(" ")

    def dot_description(self) -> str:
        """
        Returns a string representation of the node in the DOT format.

        The string includes the node's ID, half move, and board FEN.

        Returns:
            A string representation of the node in the DOT format.
        """
        return (
            "id:"
            + str(self.id)
            + " dep: "
            + str(self.half_move)
            + "\nfen:"
            + str(self.board)
        )



"""
This module contains the implementation of the CommandLineHumanMoveSelector class, which allows a human player
 to select moves through the command line interface.
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import chipiron.environments.chess_env.board as boards
from chipiron.utils import Seed
from chipiron.utils.logger import chipiron_logger

from .move_selector import MoveRecommendation
from .move_selector_types import MoveSelectorTypes

if TYPE_CHECKING:
    from chipiron.environments.chess_env.move import moveUci
    from chipiron.environments.chess_env.move.imove import moveKey


@dataclass
class CommandLineHumanPlayerArgs:
    """
    Represents the arguments for a human player that selects moves through the command line interface.
    """

    type: Literal[MoveSelectorTypes.CommandLineHuman]  # for serialization


@dataclass
class GuiHumanPlayerArgs:
    """
    Represents the arguments for a human player that selects moves through the GUI.
    """

    type: Literal[MoveSelectorTypes.GuiHuman]  # for serialization


class CommandLineHumanMoveSelector:
    """
    A move selector that allows a human player to select moves through the command line interface.
    """

    def select_move(self, board: boards.IBoard, move_seed: Seed) -> MoveRecommendation:
        """
        Selects a move based on user input through the command line interface.

        Args:
            board (boards.BoardChi): The current state of the chess board.
            move_seed (seed): The seed used for move selection.

        Returns:
            MoveRecommendation: The selected move recommendation.

        Raises:
            AssertionError: If the selected move is not a legal move.
        """
        chipiron_logger.info("Legal Moves %s", board.legal_moves)
        legal_moves_uci: list[moveUci] = []
        move: moveKey
        for move in board.legal_moves:
            legal_moves_uci.append(board.get_uci_from_move_key(move))
        chipiron_logger.info("Legal Moves %s", legal_moves_uci)

        good_move: bool = False
        move_uci: str = ""
        while not good_move:
            move_uci = input("Input your move")
            print(f"choice of move {move_uci}")
            if move_uci in legal_moves_uci:
                good_move = True
            else:
                print("Bad move, Not legal")

        assert move_uci in legal_moves_uci

        return MoveRecommendation(move=move_uci, evaluation=None)



from dataclasses import dataclass

from chipiron.players.boardevaluators.board_evaluation.board_evaluation import (
    PointOfView,
)


# this type is about the type of representation input fed to NN models
@dataclass
class ModelOutputType:
    point_of_view: PointOfView



"""
Module that contains the GUI messages that are sent to the GUI from the game.
"""

from dataclasses import dataclass

from chipiron.games.match.match_results import MatchResults


@dataclass
class MatchResultsMessage:
    """
    Represents a message containing the match results that are sent to the GUI from the game.

    Attributes:
        match_results (MatchResults): The match results object containing the details of the match.
    """

    match_results: MatchResults



"""
This module defines the AlgorithmNode class, which is a generic node used by the tree and value algorithm.
It wraps tree nodes with values, minimax computation, and exploration tools.
"""

import chess

import chipiron.environments.chess_env.board as boards
from chipiron.environments.chess_env.board.iboard import LegalMoveKeyGeneratorP
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.boardevaluators.neural_networks.input_converters.board_representation import (
    BoardRepresentation,
)
from chipiron.players.move_selector.treevalue.indices.node_indices import (
    NodeExplorationData,
)
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.node_minmax_evaluation import (
    NodeMinmaxEvaluation,
)
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode
from chipiron.players.move_selector.treevalue.nodes.tree_node import TreeNode


class AlgorithmNode:
    """
    The generic Node used by the tree and value algorithm.
    It wraps tree nodes with values, minimax computation and exploration tools
    """

    tree_node: TreeNode  # the reference to the tree node that is wrapped
    minmax_evaluation: NodeMinmaxEvaluation  # the object computing the value
    exploration_index_data: (
        NodeExplorationData | None
    )  # the object storing the information to help the algorithm decide the next nodes to explore
    board_representation: BoardRepresentation | None  # the board representation

    def __init__(
        self,
        tree_node: TreeNode,
        minmax_evaluation: NodeMinmaxEvaluation,
        exploration_index_data: NodeExplorationData | None,
        board_representation: BoardRepresentation | None,
    ) -> None:
        """
        Initializes an AlgorithmNode object.

        Args:
            tree_node (TreeNode): The tree node that is wrapped.
            minmax_evaluation (NodeMinmaxEvaluation): The object computing the value.
            exploration_index_data (NodeExplorationData | None): The object storing the information to help the algorithm decide the next nodes to explore.
            board_representation (BoardRepresentation | None): The board representation.
        """
        self.tree_node = tree_node
        self.minmax_evaluation = minmax_evaluation
        self.exploration_index_data = exploration_index_data
        self.board_representation = board_representation

    @property
    def player_to_move(self) -> chess.Color:
        """
        Returns the color of the player to move.

        Returns:
            chess.Color: The color of the player to move.
        """
        return self.tree_node.player_to_move

    @property
    def id(self) -> int:
        """
        Returns the ID of the node.

        Returns:
            int: The ID of the node.
        """
        return self.tree_node.id

    @property
    def half_move(self) -> int:
        """
        Returns the half move count.

        Returns:
            int: The half move count.
        """
        return self.tree_node.half_move

    @property
    def fast_rep(self) -> boards.boardKey:
        """
        Returns the fast representation of the node.

        Returns:
            str: The fast representation of the node.
        """
        return self.tree_node.fast_rep

    @property
    def moves_children(self) -> dict[moveKey, ITreeNode | None]:
        """
        Returns the bidirectional dictionary of moves and their corresponding child nodes.

        Returns:
            dict[IMove, ITreeNode | None]: The bidirectional dictionary of moves and their corresponding child nodes.
        """
        return self.tree_node.moves_children

    @property
    def parent_nodes(self) -> dict[ITreeNode, moveKey]:
        """
        Returns the dictionary of parent nodes of the current tree node with associated move.

        :return: A dictionary of parent nodes of the current tree node with associated move.
        """
        return self.tree_node.parent_nodes

    @property
    def board(self) -> boards.IBoard:
        """
        Returns the board.

        Returns:
            BoardChi: The board.
        """
        return self.tree_node.board

    def is_over(self) -> bool:
        """
        Checks if the game is over.

        Returns:
            bool: True if the game is over, False otherwise.
        """
        return self.minmax_evaluation.is_over()

    def add_parent(self, move: moveKey, new_parent_node: ITreeNode) -> None:
        """
        Adds a parent node.

        Args:
            move (IMove): the move that led to the node from the new_parent_node
            new_parent_node (ITreeNode): The new parent node to add.
        """
        self.tree_node.add_parent(move=move, new_parent_node=new_parent_node)

    @property
    def legal_moves(self) -> LegalMoveKeyGeneratorP:
        """
        Returns the legal move generator.

        Returns:
            chess.LegalMoveGenerator: The legal move generator.
        """
        return self.tree_node.board_.legal_moves

    @property
    def all_legal_moves_generated(self) -> bool:
        """
        Returns True if all legal moves have been generated, False otherwise.

        Returns:
            bool: True if all legal moves have been generated, False otherwise.
        """
        return self.tree_node.all_legal_moves_generated

    @all_legal_moves_generated.setter
    def all_legal_moves_generated(self, value: bool) -> None:
        """
        Sets the flag indicating if all legal moves have been generated.

        Args:
            value (bool): The value to set.
        """
        self.tree_node.all_legal_moves_generated = value

    @property
    def non_opened_legal_moves(self) -> set[moveKey]:
        """
        Returns the set of non-opened legal moves.

        Returns:
            set[IMove]: The set of non-opened legal moves.
        """
        return self.tree_node.non_opened_legal_moves

    def dot_description(self) -> str:
        """
        Returns the dot description of the node.

        Returns:
            str: The dot description of the node.
        """
        exploration_description: str = (
            self.exploration_index_data.dot_description()
            if self.exploration_index_data is not None
            else ""
        )

        return f"{self.tree_node.dot_description()}\n{self.minmax_evaluation.dot_description()}\n{exploration_description}"

    def __str__(self) -> str:
        return f"{self.__class__} id :{self.tree_node.id}"



"""
Test the creation of a Chipiron player.
"""

import random

from chipiron.players.factory import create_chipiron_player
from chipiron.scripts.chipiron_args import ImplementationArgs


def test_create_chipiron_player() -> None:
    """Test the creation of a Chipiron player."""

    # Create a Chipiron player with default arguments
    player = create_chipiron_player(
        implementation_args=ImplementationArgs(use_rust_boards=False),
        universal_behavior=True,
        random_generator=random.Random(0),
    )

    # Check if the player is created successfully
    assert player is not None
    assert player.main_move_selector is not None


if __name__ == "__main__":
    test_create_chipiron_player()



"""
Module for creating players.
"""

import queue
import random
from dataclasses import dataclass

import chess

from chipiron.players.boardevaluators.table_base.factory import (
    AnySyzygyTable,
    create_syzygy,
)
from chipiron.players.move_selector import treevalue
from chipiron.players.move_selector.treevalue.progress_monitor.progress_monitor import (
    TreeMoveLimitArgs,
)
from chipiron.players.player_args import PlayerArgs
from chipiron.players.player_ids import PlayerConfigTag
from chipiron.utils.logger import chipiron_logger

from ..environments.chess_env.board import BoardFactory, create_board_factory
from ..scripts.chipiron_args import ImplementationArgs
from ..utils.dataclass import IsDataclass
from . import move_selector
from .game_player import GamePlayer
from .player import Player
from .player_args import PlayerFactoryArgs


@dataclass
class PlayerCreationArgs:
    random_generator: random.Random
    implementation_args: ImplementationArgs
    universal_behavior: bool
    queue_progress_player: queue.Queue[IsDataclass] | None = None
    syzygy: AnySyzygyTable | None = None


def create_chipiron_player(
    implementation_args: ImplementationArgs,
    universal_behavior: bool,
    random_generator: random.Random,
    queue_progress_player: queue.Queue[IsDataclass] | None = None,
    tree_move_limit: int | None = None,
) -> Player:
    """
    Creates the chipiron champion/representative/standard/default player

    Args:
        depth: int, the depth at which computation should be made.

    Returns: the player

    """
    syzygy_table: AnySyzygyTable | None = create_syzygy(
        use_rust=implementation_args.use_rust_boards
    )

    args_player: PlayerArgs = PlayerConfigTag.CHIPIRON.get_players_args()

    if tree_move_limit is not None:
        # todo find a prettier way to do this
        assert isinstance(
            args_player.main_move_selector, treevalue.TreeAndValuePlayerArgs
        )
        assert isinstance(
            args_player.main_move_selector.stopping_criterion, TreeMoveLimitArgs
        )

        args_player.main_move_selector.stopping_criterion.tree_move_limit = (
            tree_move_limit
        )

    main_move_selector: move_selector.MoveSelector | None = (
        move_selector.create_main_move_selector(
            move_selector_instance_or_args=args_player.main_move_selector,
            syzygy=syzygy_table,
            random_generator=random_generator,
            queue_progress_player=queue_progress_player,
        )
    )

    assert main_move_selector is not None

    board_factory: BoardFactory = create_board_factory(
        use_rust_boards=implementation_args.use_rust_boards,
        use_board_modification=implementation_args.use_board_modification,
        sort_legal_moves=universal_behavior,
    )

    return Player(
        name="chipiron",
        syzygy=syzygy_table,
        main_move_selector=main_move_selector,
        board_factory=board_factory,
    )


def create_player(
    args: PlayerArgs,
    syzygy: AnySyzygyTable | None,
    random_generator: random.Random,
    implementation_args: ImplementationArgs,
    universal_behavior: bool,
    queue_progress_player: queue.Queue[IsDataclass] | None = None,
) -> Player:
    """Create a player object.

    This function creates a player object based on the provided arguments.

    Args:
        args (PlayerArgs): The arguments for creating the player.
        syzygy (AnySyzygyTable | None): The Syzygy table to be used by the player, or None if not available.
        random_generator (random.Random): The random number generator to be used by the player.

    Returns:
        Player: The created player object.
    """
    chipiron_logger.debug("Create player")
    main_move_selector: move_selector.MoveSelector = (
        move_selector.create_main_move_selector(
            move_selector_instance_or_args=args.main_move_selector,
            syzygy=syzygy,
            random_generator=random_generator,
            queue_progress_player=queue_progress_player,
        )
    )

    board_factory: BoardFactory = create_board_factory(
        use_rust_boards=implementation_args.use_rust_boards,
        use_board_modification=implementation_args.use_board_modification,
        sort_legal_moves=universal_behavior,
    )

    player: Player = Player(
        name=args.name,
        syzygy=syzygy,
        main_move_selector=main_move_selector,
        board_factory=board_factory,
    )

    return player


def create_game_player(
    player_factory_args: PlayerFactoryArgs,
    player_color: chess.Color,
    syzygy_table: AnySyzygyTable | None,
    queue_progress_player: queue.Queue[IsDataclass] | None,
    implementation_args: ImplementationArgs,
    universal_behavior: bool,
) -> GamePlayer:
    """Create a game player

    This function creates a game player using the provided player factory arguments and player color.

    Args:
        player_factory_args (PlayerFactoryArgs): The arguments for creating the player.
        player_color (chess.Color): The color of the player.

    Returns:
        GamePlayer: The created game player.
    """
    random_generator = random.Random(player_factory_args.seed)
    player: Player = create_player(
        args=player_factory_args.player_args,
        syzygy=syzygy_table,
        random_generator=random_generator,
        queue_progress_player=queue_progress_player,
        implementation_args=implementation_args,
        universal_behavior=universal_behavior,
    )
    game_player: GamePlayer = GamePlayer(player, player_color)
    return game_player



from typing import TypeAlias

from chipiron.players.boardevaluators.neural_networks.models.multi_layer_perceptron import (
    MultiLayerPerceptronArgs,
)
from chipiron.players.boardevaluators.neural_networks.models.transformer_one import (
    TransformerArgs,
)

NNModelTypeArgs: TypeAlias = MultiLayerPerceptronArgs | TransformerArgs



"""
This module defines the different types of board evaluators used in the tree value calculation.
"""

from enum import Enum


class BoardEvalTypes(str, Enum):
    """
    Enumeration of argument types for different board evaluators.

    Attributes:
        NEURAL_NET_BOARD_EVAL: Argument type for neural network-based board evaluation.
        STOCKFISH_BOARD_EVAL: Argument type for Stockfish engine-based board evaluation.
        TABLE_BASE: Argument type for tablebase-based board evaluation.
        BASIC_EVALUATION: Argument type for basic/static board evaluation.
    """

    NEURAL_NET_BOARD_EVAL = "neural_network"
    STOCKFISH_BOARD_EVAL = "stockfish"
    TABLE_BASE_EVAL = "table_base"
    BASIC_EVALUATION_EVAL = "basic_evaluation"



"""
Path utilities for the Chipiron project.

This module provides centralized path definitions to ensure consistent
path handling across the entire project.
"""

import os
from pathlib import Path

# Define the project root directory
# This file is at: chipiron/utils/path_variables.py
# So we go up 2 levels to reach project root
PROJECT_ROOT = Path(__file__).parents[2]  # Go up from chipiron/utils/ to project root

# Load .env file if it exists
env_file = PROJECT_ROOT / ".env"
if env_file.exists():
    with open(env_file, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith("#"):
                if "=" in line:
                    key, value = line.split("=", 1)
                    os.environ.setdefault(key, value)


def get_env_path(env_var: str, default: str) -> Path:
    """Get a path from environment variable or use default relative to PROJECT_ROOT."""
    env_value = os.getenv(env_var, default)
    if os.path.isabs(env_value):
        return Path(env_value)
    return PROJECT_ROOT / env_value


# Common directory paths
EXTERNAL_DATA_DIR = get_env_path("EXTERNAL_DATA_DIR", "external_data")
LICHESS_PGN_DIR = get_env_path("LICHESS_PGN_DIR", "external_data/lichess_pgn")
SYZYGY_TABLES_DIR = get_env_path("SYZYGY_TABLES_DIR", "external_data/syzygy-tables")
STOCKFISH_DIR = get_env_path("STOCKFISH_DIR", "external_data/stockfish")
GUI_DIR = get_env_path("GUI_DIR", "external_data/gui")
SCRIPTS_DIR = PROJECT_ROOT / "chipiron" / "scripts"

# Specific file paths
LICHESS_PGN_FILE = get_env_path(
    "LICHESS_PGN_FILE",
    "external_data/lichess_pgn/lichess_db_standard_rated_2015-03.pgn",
)
STOCKFISH_BINARY_PATH = get_env_path(
    "STOCKFISH_BINARY_PATH",
    "external_data/stockfish/stockfish/stockfish-ubuntu-x86-64-avx2",
)

# MLflow paths (use environment variables with defaults)
ML_FLOW_URI_PATH = os.getenv(
    "ML_FLOW_URI_PATH",
    f"sqlite:///{PROJECT_ROOT}/chipiron/scripts/default_output_folder/mlflow_data/mlruns.db",
)
ML_FLOW_URI_PATH_TEST = os.getenv(
    "ML_FLOW_URI_PATH_TEST",
    f"sqlite:///{PROJECT_ROOT}/chipiron/scripts/default_output_folder/mlflow_data/mlruns_test.db",
)


def main() -> None:
    """
    Print all defined paths in absolute form for debugging and verification.
    """
    print("Chipiron Project Paths (with .env integration):")
    print("=" * 50)
    print(f"PROJECT_ROOT:        {PROJECT_ROOT.absolute()}")
    print()
    print("Directory Paths:")
    print(f"EXTERNAL_DATA_DIR:   {EXTERNAL_DATA_DIR.absolute()}")
    print(f"LICHESS_PGN_DIR:     {LICHESS_PGN_DIR.absolute()}")
    print(f"SYZYGY_TABLES_DIR:   {SYZYGY_TABLES_DIR.absolute()}")
    print(f"STOCKFISH_DIR:       {STOCKFISH_DIR.absolute()}")
    print(f"GUI_DIR:             {GUI_DIR.absolute()}")
    print(f"SCRIPTS_DIR:         {SCRIPTS_DIR.absolute()}")
    print()
    print("File Paths:")
    print(f"LICHESS_PGN_FILE:    {LICHESS_PGN_FILE.absolute()}")
    print(f"STOCKFISH_BINARY_PATH: {STOCKFISH_BINARY_PATH.absolute()}")
    print()
    print("MLflow Paths:")
    print(f"ML_FLOW_URI_PATH:    {ML_FLOW_URI_PATH}")
    print(f"ML_FLOW_URI_PATH_TEST: {ML_FLOW_URI_PATH_TEST}")
    print()
    print("Environment Variables Used:")
    env_vars = [
        "EXTERNAL_DATA_DIR",
        "LICHESS_PGN_DIR",
        "SYZYGY_TABLES_DIR",
        "STOCKFISH_DIR",
        "GUI_DIR",
        "LICHESS_PGN_FILE",
        "STOCKFISH_BINARY_PATH",
    ]
    for var in env_vars:
        env_value = os.getenv(var, "NOT SET")
        print(f"{var}: {env_value}")
    print()
    print("Path Existence Check:")
    print(f"PROJECT_ROOT exists:     {PROJECT_ROOT.exists()}")
    print(f"EXTERNAL_DATA_DIR exists: {EXTERNAL_DATA_DIR.exists()}")
    print(f"GUI_DIR exists:          {GUI_DIR.exists()}")
    print(f"LICHESS_PGN_FILE exists:  {LICHESS_PGN_FILE.exists()}")


if __name__ == "__main__":
    main()



"""
Module containing the class GamePlaying
"""

from dataclasses import dataclass
from enum import Enum


class PlayingStatus(Enum):
    """
    Class containing the possible playing status of a game.
    """

    PLAY = 1
    PAUSE = 0


@dataclass
class GamePlayingStatus:
    """
    Object containing the playing status of a game
    and the board.
    """

    _status: PlayingStatus = PlayingStatus.PLAY

    @property
    def status(self) -> PlayingStatus:
        """
        Get the current playing status of the game.

        Returns:
            PlayingStatus: The current playing status.
        """
        return self._status

    @status.setter
    def status(self, new_status: PlayingStatus) -> None:
        """
        Set the playing status of the game.

        Args:
            new_status (PlayingStatus): The new playing status.
        """
        self._status = new_status

    def play(self) -> None:
        """
        Set the playing status to 'PLAY'.
        """
        self.status = PlayingStatus.PLAY

    def pause(self) -> None:
        """
        Set the playing status to 'PAUSE'.
        """
        self.status = PlayingStatus.PAUSE

    def is_paused(self) -> bool:
        """
        Check if the game is currently paused.

        Returns:
            bool: True if the game is paused, False otherwise.
        """
        return self.status == PlayingStatus.PAUSE

    def is_play(self) -> bool:
        """
        Check if the game is currently playing.

        Returns:
            bool: True if the game is playing, False otherwise.
        """
        return self.status == PlayingStatus.PLAY



"""
This module contains a factory to create board representations.
"""

from typing import Any

from .factory import RepresentationFactory
from .ModelInputRepresentationType import InternalTensorRepresentationType
from .rep_364 import create_from_board as create_from_board_364_no_bug
from .rep_364 import (
    create_from_board_and_from_parent as create_from_board_and_from_parent_364_no_bug,
)
from .rep_364_bug import create_from_board as create_from_board_364_bug
from .rep_364_bug import (
    create_from_board_and_from_parent as create_from_board_and_from_parent_364_bug,
)


def create_board_representation_factory(
    internal_tensor_representation_type: InternalTensorRepresentationType,
) -> RepresentationFactory | None:
    """
    Create a board representation based on the given string.

    Args:
        internal_tensor_representation_type (InternalTensorRepresentationType): The string representing the board representation.

    Returns:
        Representation364Factory | None: The created board representation object, or None if the string is 'no'.

    Raises:
        Exception: If the string is not '364'.

    """
    board_representation_factory: RepresentationFactory[Any] | None
    match internal_tensor_representation_type:
        case InternalTensorRepresentationType.BUG364:
            board_representation_factory = RepresentationFactory(
                create_from_board=create_from_board_364_bug,
                create_from_board_and_from_parent=create_from_board_and_from_parent_364_bug,
            )
        case InternalTensorRepresentationType.NOBUG364:
            board_representation_factory = RepresentationFactory(
                create_from_board=create_from_board_364_no_bug,
                create_from_board_and_from_parent=create_from_board_and_from_parent_364_no_bug,
            )
        case InternalTensorRepresentationType.NO:
            board_representation_factory = None
        case _:
            raise ValueError(
                f"trying to create {internal_tensor_representation_type} in file {__name__}"
            )

    return board_representation_factory



"""
Tests for the board representation.
"""

from copy import deepcopy
from typing import TYPE_CHECKING, Any

import chess
import pytest
import torch

from chipiron.environments.chess_env.board import (
    BoardModificationP,
    IBoard,
    create_board,
)
from chipiron.environments.chess_env.board.utils import FenPlusHistory, bitboard_rotate
from chipiron.players.boardevaluators.neural_networks.input_converters.ModelInputRepresentationType import (
    InternalTensorRepresentationType,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.representation_factory_factory import (
    create_board_representation_factory,
)

if TYPE_CHECKING:
    from chipiron.environments.chess_env.move.imove import moveKey
    from chipiron.players.boardevaluators.neural_networks.input_converters.board_representation import (
        BoardRepresentation,
        Representation364,
    )
    from chipiron.players.boardevaluators.neural_networks.input_converters.factory import (
        RepresentationFactory,
    )


@pytest.mark.parametrize(("use_rust_boards"), (True, False))
@pytest.mark.parametrize(("board_representation_factory_type"), ["364_no_bug"])
def test_representation(
    use_rust_boards: bool,
    board_representation_factory_type: InternalTensorRepresentationType,
) -> None:
    """Test the board representation factory.

    Args:
        use_rust_boards (bool): Whether to use Rust boards.
        board_representation_factory_type (InternalTensorRepresentationType): The type of board representation factory to use.
    """
    board: IBoard = create_board(
        use_rust_boards=use_rust_boards,
        use_board_modification=True,
        fen_with_history=FenPlusHistory(current_fen=chess.STARTING_FEN),
    )

    representation_factory: RepresentationFactory[Any] | None = (
        create_board_representation_factory(
            internal_tensor_representation_type=board_representation_factory_type
        )
    )
    assert representation_factory is not None

    parent_node_board_representation: Representation364 = (
        representation_factory.create_from_board(board=board)
    )

    all_moves_keys_chi: list[moveKey] = board.legal_moves.get_all()
    board_modification: BoardModificationP | None = board.play_move_key(
        move=all_moves_keys_chi[0]
    )

    assert board_modification is not None
    direct_rep: BoardRepresentation = representation_factory.create_from_board(
        board=board
    )
    rep_from_parents: BoardRepresentation = (
        representation_factory.create_from_board_and_from_parent(
            board=board,
            board_modifications=board_modification,
            parent_node_board_representation=parent_node_board_representation,
        )
    )

    assert direct_rep == rep_from_parents


@pytest.mark.parametrize(("board_representation_factory_type"), ["364_no_bug"])
@pytest.mark.parametrize(("use_rust_boards"), (True, False))
@pytest.mark.parametrize(("use_board_modification"), (True, False))
def test_representation364(
    use_rust_boards: bool,
    board_representation_factory_type: InternalTensorRepresentationType,
    use_board_modification: bool,
) -> None:
    # 'rnb2bnr/ppp2ppp/2k3q1/8/8/1Q3K2/PPP2PPP/RNB2BNR w - - 0 1'
    # '8/ppp2ppp/2k3q1/8/8/1Q3K2/PPP2PPP/8 w - - 0 1'
    board_one: IBoard = create_board(
        use_rust_boards=use_rust_boards,
        use_board_modification=use_board_modification,
        fen_with_history=FenPlusHistory(
            current_fen="rnb2bnr/ppp2ppp/2k3q1/8/8/1Q3K2/PPP2PPP/RNB2BNR w - - 0 1"
        ),
    )

    board_two: IBoard = create_board(
        use_rust_boards=use_rust_boards,
        use_board_modification=use_board_modification,
        fen_with_history=FenPlusHistory(
            current_fen="rnb2bnr/ppp2ppp/2k3q1/8/8/1Q3K2/PPP2PPP/RNB2BNR b - - 0 1"
        ),
    )

    representation_factory: RepresentationFactory[Any] | None = (
        create_board_representation_factory(
            internal_tensor_representation_type=board_representation_factory_type
        )
    )
    assert representation_factory is not None

    assert board_one.occupied == bitboard_rotate(board_two.occupied)

    board_representation_one_copy: Representation364 = deepcopy(
        representation_factory.create_from_board(board=board_one)
    )
    board_representation_two_copy: Representation364 = deepcopy(
        representation_factory.create_from_board(board=board_two)
    )

    board_representation_one: Representation364
    board_representation_two: Representation364
    if use_board_modification:
        board_modification_one: BoardModificationP | None = board_one.play_move_uci(
            move_uci="a2a3"
        )
        board_modification_two: BoardModificationP | None = board_two.play_move_uci(
            move_uci="h7h6"
        )
        assert board_modification_one is not None
        assert board_modification_two is not None
        assert board_one.occupied == bitboard_rotate(board_two.occupied)
        board_representation_one = (
            representation_factory.create_from_board_and_from_parent(
                board=board_one,
                board_modifications=board_modification_one,
                parent_node_board_representation=board_representation_one_copy,
            )
        )
        board_representation_two = (
            representation_factory.create_from_board_and_from_parent(
                board=board_two,
                board_modifications=board_modification_two,
                parent_node_board_representation=board_representation_two_copy,
            )
        )
    else:
        board_representation_one = representation_factory.create_from_board(
            board=board_one
        )
        board_representation_two = representation_factory.create_from_board(
            board=board_two
        )

    inputs_one = board_representation_one.get_evaluator_input(board_one.turn)

    inputs_two = board_representation_two.get_evaluator_input(board_two.turn)

    assert torch.equal(inputs_one, inputs_two)


if __name__ == "__main__":
    use_rusty_board: bool
    board_representation_factory_type: InternalTensorRepresentationType
    use_board_modification: bool
    for use_rusty_board in [True, False]:
        for board_representation_factory_type in [
            InternalTensorRepresentationType.NOBUG364
        ]:
            test_representation(
                use_rust_boards=use_rusty_board,
                board_representation_factory_type=board_representation_factory_type,
            )
            for use_board_modification in [True, False]:
                test_representation364(
                    use_rust_boards=use_rusty_board,
                    board_representation_factory_type=board_representation_factory_type,
                    use_board_modification=use_board_modification,
                )



"""
Module for the ChiNN class
"""

import sys
import traceback

import torch
import torch.nn as nn

from chipiron.utils import path
from chipiron.utils.logger import chipiron_logger
from chipiron.utils.small_tools import resolve_package_path


class ChiNN(nn.Module):
    """
    The Generic Neural network class of chipiron
    """

    def __init__(self) -> None:
        """
        Initializes an instance of the ChiNN class.
        """
        super().__init__()

    def __getstate__(self) -> dict[str, object]:
        """
        Get the state of the object for pickling.

        Returns:
            dict: The state dictionary of the object.
        """
        state = self.__dict__.copy()
        return state

    def init_weights(self) -> None:
        """
        Initialize the weights of the model.
        """
        pass

    def load_weights_from_file(self, path_to_param_file: path) -> None:
        """
        Loads the neural network weights from a file or initializes them if the file doesn't exist.

        Args:
            path_to_param_file (str): The path to the parameter file.
            authorisation_to_create_file (bool): Flag indicating whether the program has authorization to create a new file.

        Returns:
            None
        """
        chipiron_logger.info("load_or_init_weights from %s", path_to_param_file)
        try:  # load
            resolved_path = resolve_package_path(str(path_to_param_file))
            with open(resolved_path, "rb") as fileNNR:
                chipiron_logger.info(
                    "loading the existing param file %s", resolved_path
                )
                if torch.cuda.is_available():
                    self.load_state_dict(torch.load(fileNNR))
                else:
                    self.load_state_dict(
                        torch.load(fileNNR, map_location=torch.device("cpu"))
                    )

        except EnvironmentError:  # init
            # Print the full traceback to stderr
            traceback.print_exc()

            resolved_path = resolve_package_path(str(path_to_param_file))
            chipiron_logger.error("no file %s at %s", path_to_param_file, resolved_path)
            sys.exit(
                f"Error: no NN weights file and no rights to create it for file {path_to_param_file}"
            )

    def log_readable_model_weights_to_file(self, file_path: str) -> None:
        """Logs the model weights to a file in a human-readable format.

        Args:
            file_path (str): The path to the file where the weights will be logged.

        Raises:
            Exception: If the logging fails.
        """
        raise NotImplementedError("not implemented in base class")
